/*
 * Copyright (c) 2003-2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: resource.c,v 1.33 2011/08/02 21:02:12 ca Exp $")
#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/qmgr.h"
#include "sm/qmgr-int.h"
#include "qm_throttle.h"
#include "qmgr.h"
#include "log.h"

/*
**  QM_FS_GETFREE -- call *fs_getfree() for each "db"
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code
**
**	Locking:
**		qmgr_ctx is locked/unlocked as requested by locktype
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
qm_fs_getfree(qmgr_ctx_P qmgr_ctx, thr_lock_T locktype)
{
	sm_ret_T ret;
	int r;

	SM_IS_QMGR_CTX(qmgr_ctx);
	ret = SM_SUCCESS;

	if (thr_lock_it(locktype)) {
		r = pthread_mutex_lock(&qmgr_ctx->qmgr_mutex);
		SM_LOCK_OK(r);
		if (r != 0) {
			QM_LEV_DPRINTFC(QDC_RSRC, 0, (QM_DEBFP, "sev=ERROR, func=qm_fs_getfree, lock=%d\n", r));
			return sm_error_temp(SM_EM_Q, r);
		}
	}

	ret = cdb_fs_getfree(qmgr_ctx->qmgr_cdb_fsctx, &qmgr_ctx->qmgr_cdb_kbfree);
	QM_LEV_DPRINTFC(QDC_RSRC, 3, (QM_DEBFP, "func=qm_fs_getfree, cdb_kbfree=%lu, ret=%r\n", qmgr_ctx->qmgr_cdb_kbfree, ret));
	if (sm_is_success(ret)) {
		qmgr_ctx->qmgr_usage[QMGR_RFL_CDB_I] =
			DISK_USAGE(qmgr_ctx->qmgr_cdb_kbfree, qmgr_ctx);
		QMGR_SET_USE_MAX(QMGR_RFL_CDB_I);
	}
	ret = edb_fs_getfree(qmgr_ctx->qmgr_edb_fsctx, &qmgr_ctx->qmgr_edb_kbfree);
	QM_LEV_DPRINTFC(QDC_RSRC, 3, (QM_DEBFP, "func=qm_fs_getfree, edb_kbfree=%lu, ret=%r\n", qmgr_ctx->qmgr_edb_kbfree, ret));
	if (sm_is_success(ret)) {
		qmgr_ctx->qmgr_usage[QMGR_RFL_EDB_I] =
			DISK_USAGE(qmgr_ctx->qmgr_edb_kbfree, qmgr_ctx);
		QMGR_SET_USE_MAX(QMGR_RFL_EDB_I);
	}
	ret = ibdb_fs_getfree(qmgr_ctx->qmgr_ibdb, &qmgr_ctx->qmgr_ibdb_kbfree);
	QM_LEV_DPRINTFC(QDC_RSRC, 3, (QM_DEBFP, "func=qm_fs_getfree, ibdb_kbfree=%lu, ret=%r\n", qmgr_ctx->qmgr_ibdb_kbfree, ret));
	if (sm_is_success(ret)) {
		qmgr_ctx->qmgr_usage[QMGR_RFL_IBD_I] =
			DISK_USAGE(qmgr_ctx->qmgr_ibdb_kbfree, qmgr_ctx);
		QMGR_SET_USE_MAX(QMGR_RFL_IBD_I);
	}

	if (thr_unl_always(locktype)) {
		r = pthread_mutex_unlock(&qmgr_ctx->qmgr_mutex);
		SM_ASSERT(0 == r);
		if (r != 0 && sm_is_success(ret))
			ret = sm_error_perm(SM_EM_Q, r);
	}
	return ret;
}


/*
**  QM_COMP_RESOURCE -- compute total resource usage
**	result stored in qmgr_ctx->qmgr_total_usage
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code
**
**	Called by:
**		qss_control()
**		qss_unthrottle()
**
**	Locking:
**		qmgr_ctx is locked/unlocked as requested by locktype
**		check unlocking! XXX
**
**	Last code review: 2005-04-21 23:54:01
**	Last code change:
*/

sm_ret_T
qm_comp_resource(qmgr_ctx_P qmgr_ctx, thr_lock_T locktype)
{
	sm_ret_T ret;
	int r;
	uint u;
	uint total, sum;

	SM_IS_QMGR_CTX(qmgr_ctx);
	ret = SM_SUCCESS;

	if (thr_lock_it(locktype)) {
		r = pthread_mutex_lock(&qmgr_ctx->qmgr_mutex);
		SM_LOCK_OK(r);
		if (r != 0) {
			QM_LEV_DPRINTFC(QDC_RSRC, 0, (QM_DEBFP, "sev=ERROR, func=qm_comp_resource, lock=%d\n", r));
			return sm_error_temp(SM_EM_Q, r);
		}
	}

	ret = qm_fs_getfree(qmgr_ctx, THR_NO_LOCK);

	total = QMGR_R_USE_NONE;
	sum = 0;
	for (u = 0; u <= QMGR_RFL_LAST_I; u++) {
		if (qmgr_ctx->qmgr_usage[u] >= qmgr_ctx->qmgr_upper[u]) {
			total = QMGR_R_USE_FULL;
			break;
		}
		if (qmgr_ctx->qmgr_usage[u] > qmgr_ctx->qmgr_lower[u]) {
			sum += 100 *
				(qmgr_ctx->qmgr_usage[u] - qmgr_ctx->qmgr_lower[u]) /
				(qmgr_ctx->qmgr_upper[u] - qmgr_ctx->qmgr_lower[u]);
		}
	}
	if (QMGR_R_USE_NONE == total)
		total = sum / (QMGR_RFL_LAST_I + 1);
	qmgr_ctx->qmgr_total_usage = total;

	if (thr_unl_always(locktype)) {
		r = pthread_mutex_unlock(&qmgr_ctx->qmgr_mutex);
		SM_ASSERT(0 == r);
		if (r != 0 && sm_is_success(ret))
			ret = sm_error_perm(SM_EM_Q, r);
	}
	return ret;
}

/*
**  QM_RESOURCE -- change resource usage
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		direction -- up/down/any direction
**		use -- fullness of cache/DB/resource (0..100)
**		resource -- which resource?
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code
**
**	Locking:
**		qmgr_ctx is locked if requested
**
**	Last code review: 2005-04-23 21:52:26
**	Last code change:
*/

sm_ret_T
qm_resource(qmgr_ctx_P qmgr_ctx, int direction, uint use, uint resource, thr_lock_T locktype)
{
	sm_ret_T ret;
	int r;

	SM_IS_QMGR_CTX(qmgr_ctx);
	SM_REQUIRE(resource <= QMGR_RFL_LAST_I);
	ret = SM_SUCCESS;

	QM_LEV_DPRINTFC(QDC_RSRC, 6, (QM_DEBFP, "func=qm_resource, dir=%d, use=%u, resource=%u, flags=%#x\n", direction, use, resource, qmgr_ctx->qmgr_rflags));

	if (thr_lock_it(locktype)) {
		r = pthread_mutex_lock(&qmgr_ctx->qmgr_mutex);
		SM_LOCK_OK(r);
		if (r != 0) {
			sm_log_write(qmgr_ctx->qmgr_lctx,
				QM_LCAT_CONTROL, QM_LMOD_CONTROL,
				SM_LOG_CRIT, 4,
				"sev=CRIT, func=qm_resource, lock=%d", r);
			return sm_error_temp(SM_EM_Q, r);
		}
	}

	/*
	**  Throttle the system iff
	**  - asked for and
	**  - the use of the resource exceeds the upper threshold and
	**  - the use of the resource exceeds the last recorded usage
	**
	**  else unthrottle the system iff
	**  - asked for and
	**  - the use of the resource is below the lower threshold and
	**  - the use of the resource is below the last recorded usage and
	**  - the resource is flagged as being exceeded
	**
	**   in that case unthrottle the SMTPS servers iff
	**   - there is no resource shortage at all
	*/

	if (qmgr_is_throttle(direction)
	    && use > qmgr_ctx->qmgr_upper[resource]
	    && use > qmgr_ctx->qmgr_usage[resource])
	{
		QMGR_SET_RFLAG_I(qmgr_ctx, resource);
		QM_LEV_DPRINTFC(QDC_RSRC, 2, (QM_DEBFP, "func=qm_resource, status=throttle, use=%u, resource=%d, flags=%#x\n", use, resource, qmgr_ctx->qmgr_rflags));
	}
	else if (qmgr_is_un_throttle(direction)
		 && use < qmgr_ctx->qmgr_lower[resource]
		 && use < qmgr_ctx->qmgr_usage[resource]
		 && QMGR_IS_RFLAG_I(qmgr_ctx, resource)
		)
	{
		QMGR_CLR_RFLAG_I(qmgr_ctx, resource);
		QM_LEV_DPRINTFC(QDC_RSRC, 2, (QM_DEBFP, "func=qm_resource, status=unthrottle, use=%u, resource=%d, flags=%#x\n", use, resource, qmgr_ctx->qmgr_rflags));

		if (QMGR_RFL_IS_NO_RSR(qmgr_ctx)) {
			sm_log_write(qmgr_ctx->qmgr_lctx,
				QM_LCAT_CONTROL, QM_LMOD_CONTROL,
				SM_LOG_INFO, 9,
				"sev=INFO, func=qm_resource, status=unthrottle, use=%u, resource=%s, flags=%#x"
				, use, qm_rn[resource], qmgr_ctx->qmgr_rflags);
			ret = qss_wakeup(qmgr_ctx, THR_NO_LOCK);
			if (sm_is_err(ret))
				QM_LEV_DPRINTFC(QDC_RSRC, 1, (QM_DEBFP, "sev=ERROR, func=qm_resource, qss_wakeup=%r\n", ret));
		}
	}
	qmgr_ctx->qmgr_usage[resource] = use;
	QMGR_SET_USE_MAX(resource);
	if (thr_unl_always(locktype)) {
		r = pthread_mutex_unlock(&qmgr_ctx->qmgr_mutex);
		SM_ASSERT(0 == r);
		if (r != 0 && sm_is_success(ret))
			ret = sm_error_perm(SM_EM_Q, r);
	}
	return ret;
}

/*
**  QM_FS_CHK -- check fs usage, unthrottle if possible
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		chkonly -- only check usage
**		locktype -- kind of locking
**
**	Returns:
**		usual sm_error code
**		if "disk full": SM_NEEDMORE to indicate to another check is needed
**
**	Locking:
**		qmgr_ctx is locked/unlocked as requested by locktype
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
qm_fs_chk(qmgr_ctx_P qmgr_ctx, bool chkonly, thr_lock_T locktype)
{
	sm_ret_T ret;
	int r;
	uint u;
	static uint fschks[] = { QMGR_RFL_CDB_I, QMGR_RFL_EDB_I, QMGR_RFL_IBD_I };
	bool full;

	SM_IS_QMGR_CTX(qmgr_ctx);
	ret = SM_SUCCESS;

	if (thr_lock_it(locktype)) {
		r = pthread_mutex_lock(&qmgr_ctx->qmgr_mutex);
		SM_LOCK_OK(r);
		if (r != 0) {
			QM_LEV_DPRINTFC(QDC_RSRC, 0, (QM_DEBFP, "sev=ERROR, func=qm_fs_chk, lock=%d\n", r));
			return sm_error_temp(SM_EM_Q, r);
		}
	}

	if (chkonly) {
		full = false;
		for (u = 0; u < SM_ARRAY_SIZE(fschks); u++) {
			full |= qmgr_ctx->qmgr_usage[fschks[u]] >= qmgr_ctx->qmgr_upper[fschks[u]];
		}
	}
	else {
		full = false;
		for (u = 0; u < SM_ARRAY_SIZE(fschks); u++) {
			if (qmgr_ctx->qmgr_usage[fschks[u]] >= qmgr_ctx->qmgr_upper[fschks[u]])
			{
				ret = qm_fs_getfree(qmgr_ctx, THR_NO_LOCK);
				full = true;
				break;
			}
		}

		if (full) {
			for (u = 0; u < SM_ARRAY_SIZE(fschks); u++) {
				ret = qm_resource(qmgr_ctx, QMGR_UN_THROTTLE,
					qmgr_ctx->qmgr_usage[fschks[u]], fschks[u], THR_NO_LOCK);
			}
			full = false;
			for (u = 0; u < SM_ARRAY_SIZE(fschks); u++) {
				full |= qmgr_ctx->qmgr_usage[fschks[u]] >= qmgr_ctx->qmgr_upper[fschks[u]];
			}
		}
	}

	if (thr_unl_always(locktype)) {
		r = pthread_mutex_unlock(&qmgr_ctx->qmgr_mutex);
		SM_ASSERT(0 == r);
		if (r != 0 && sm_is_success(ret))
			ret = sm_error_perm(SM_EM_Q, r);
	}
	return sm_is_success(ret) ? (full ? SM_NEEDMORE : SM_SUCCESS) : ret;
}
