/*
 * Copyright (c) 2002-2006, 2011 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: sched.c,v 1.384 2014/01/28 03:12:37 ca Exp $")
#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/io.h"
#include "sm/rcb.h"
#include "sm/reccom.h"
#include "sm/da.h"
#include "sm/das.h"
#include "sm/qmgr.h"
#include "sm/qmgr-int.h"
#include "sm/misc.h"
#include "qmgr.h"
#include "sm/aqrdq.h"
#include "log.h"

/*
**  Which access methods to recipients does the scheduler need?
**  DA/Host: for session reuse.
**  SMTPS TA-Id: to find recipients for the same TA (that could be
**	be delivered in one TA if the DA/Host is the same).
**
**  Which other (recipient) data does the scheduler need?
**  Number of recipients/TAs destined for the same DA/Host.
*/

/*
**  Can both recipients be delivered in the same transaction?
**  They are from the same incoming transaction, but do they share
**  same delivery information (MX piggypacking?)
**  Bounces can't be piggypacked: they have content in the rcpt itself.
**  Other restrictions? e.g., TLS requirements, client IP address, ...
**  Should this check AQ_TA_IS_FLAG(aq_ta, AQ_TA_FL_VERP)?
*/

/*
 *  Note: this macro only works for
 *  aq_rcpt->aqr_ss_ta_id == aq_rcpt_nxt->aqr_ss_ta_id
 *  use SAME_TA() if that's not guaranteed.
 */
#define SAME_TRANSACTION(aq_rcpt, aq_rcpt_nxt)	\
	(IPADDR_EQ((aq_rcpt)->aqr_addrs[(aq_rcpt)->aqr_addr_cur].aqra_ip, \
	 (aq_rcpt_nxt)->aqr_addrs[(aq_rcpt_nxt)->aqr_addr_cur].aqra_ip) \
	&& (aq_rcpt)->aqr_da_idx == (aq_rcpt_nxt)->aqr_da_idx \
	&& (aq_rcpt)->aqr_port == (aq_rcpt_nxt)->aqr_port \
	&& (aq_rcpt)->aqr_owner_idx == (aq_rcpt_nxt)->aqr_owner_idx \
	&& !AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_DSN) \
	&& !AQR_IS_FLAG(aq_rcpt_nxt, AQR_FL_IS_DSN) \
	&& !AQR_IS_FLAG(aq_rcpt, AQR_FL_HAS_VERP) \
	&& !AQR_IS_FLAG(aq_rcpt_nxt, AQR_FL_HAS_VERP) \
	)
#define SAME_TA(aq_rcpt, aq_rcpt_nxt)	\
		(SESSTA_EQ(aq_rcpt->aqr_ss_ta_id, aq_rcpt_nxt->aqr_ss_ta_id) && \
		SAME_TRANSACTION(aq_rcpt, aq_rcpt_nxt))

/* primitive version of setting an error value if it isn't set already */
#define SM_SET_RET(res, ret) do {				\
		if (sm_is_err(ret) && !sm_is_err(res))	\
			(res) = (ret);			\
	} while (0)

/*
**  QM_AR_ACTSCHED -- Decide whether to activate scheduler after getting
**	result from AR.
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_rcpt -- AQ recipient
**		pnotify_sched -- (pointer to) notify scheduler? (output)
**
**	Returns:
**		usual sm_error code; only (un)lock
**
**	Last code review: 2005-04-04 20:39:02
**	Last code change: 2005-04-04 20:37:48
*/

sm_ret_T
qm_ar_actsched(qmgr_ctx_P qmgr_ctx, aq_rcpt_P aq_rcpt, bool *pnotify_sched)
{
	sm_ret_T ret;
	int r;
	occ_entry_P occ_entry;

	SM_IS_QMGR_CTX(qmgr_ctx);
	SM_IS_AQ_RCPT(aq_rcpt);
	SM_REQUIRE(pnotify_sched != NULL);

	/* default: activate scheduler */
	*pnotify_sched = true;
	ret = SM_SUCCESS;
	if (!AQR_IS_FLAG(aq_rcpt, AQR_FL_ARF) && aq_rcpt->aqr_addrs != NULL) {
		occ_entry = NULL;
		SM_ASSERT(aq_rcpt->aqr_addr_cur < aq_rcpt->aqr_addr_max);
		ret = occ_entry_find(qmgr_ctx->qmgr_occ_ctx->occx_ht,
			&AQR_CUR_DEST(aq_rcpt),
			&occ_entry, &qmgr_ctx->qmgr_occ_ctx->occx_mutex,
			THR_LOCK_UNLERR);
		if (SM_SUCCESS == ret && occ_entry != NULL) {
			if (occ_entry->occe_open_se >= occ_entry->occe_cur_conc
			    && occ_entry->occe_open_se == occ_entry->occe_open_ta)
			{
#if 1
				QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP,
					"func=qm_ar_actsched, open_se=%d, cur_conc=%d, open_se=%d, open_ta=%d\n"
					, occ_entry->occe_open_se, occ_entry->occe_cur_conc
					, occ_entry->occe_open_se, occ_entry->occe_open_ta));
#endif /* 0 */
				*pnotify_sched = false;
				OCCE_SET_FLAG(occ_entry, OCCE_FL_SE_WAIT);
			}
			r = smthread_mutex_unlock(&qmgr_ctx->qmgr_occ_ctx->occx_mutex);
			SM_ASSERT(0 == r);
			if (r != 0)
				ret = sm_error_perm(SM_EM_Q_SCHED, r);
		}
		else
			ret = SM_SUCCESS;
	}
#if 0
	/* else: scheduler might be waiting for result, see below: */
	if (!((0 == aq_ta->aqt_rcpts_ar)
	      || (aq_ta->aqt_rcpts_ar <= (aq_ta->aqt_rcpts_inaq >> 1)
		  && (aq_ctx->aq_entries < (aq_ctx->aq_t_da >> 4)))
#endif /* 0 */

	return ret;
}

/*
**  QM_FIND_RDQ -- find an IP for which the OC limits are not exceeded
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_rcpt -- AQ recipient
**
**	Returns:
**		usual sm_error code;
**
**	Side Effects: may modify aq_rcpt->aqr_addr_cur
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
qm_find_rdq(qmgr_ctx_P qmgr_ctx, aq_rcpt_P aq_rcpt)
{
	sm_ret_T ret;
	int r;
	uint addr_initial;
	bool gotit;
	occ_entry_P occ_entry;

	SM_IS_QMGR_CTX(qmgr_ctx);
	SM_IS_AQ_RCPT(aq_rcpt);

	if (AQR_IS_FLAG(aq_rcpt, AQR_FL_ARF) || NULL == aq_rcpt->aqr_addrs)
		return SM_NOTDONE;

	gotit = false;
	occ_entry = NULL;
	addr_initial = aq_rcpt->aqr_addr_cur;
	/* should we stop before "backupmx"? */
	while (aq_rcpt->aqr_addr_cur < aq_rcpt->aqr_addr_max) {
		ret = occ_entry_find(qmgr_ctx->qmgr_occ_ctx->occx_ht,
			&AQR_CUR_DEST(aq_rcpt),
			&occ_entry, &qmgr_ctx->qmgr_occ_ctx->occx_mutex,
			THR_LOCK_UNLERR);
		if (NULL == occ_entry)
			return SM_SUCCESS;
		if (SM_SUCCESS == ret && occ_entry != NULL) {
			if (!(occ_entry->occe_open_se >= occ_entry->occe_cur_conc
			      && occ_entry->occe_open_se == occ_entry->occe_open_ta)) {
				gotit = true;
				QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP,
					"func=qm_find_rdq, open_se=%d, cur_conc=%d, open_se=%d, open_ta=%d\n"
					, occ_entry->occe_open_se, occ_entry->occe_cur_conc
					, occ_entry->occe_open_se, occ_entry->occe_open_ta));
			}
			r = smthread_mutex_unlock(&qmgr_ctx->qmgr_occ_ctx->occx_mutex);
			SM_ASSERT(0 == r);
			if (r != 0)
				ret = sm_error_perm(SM_EM_Q_SCHED, r);
			if (gotit)
				return ret;
		}
		++aq_rcpt->aqr_addr_cur;
	}
	aq_rcpt->aqr_addr_cur = addr_initial;
	return sm_err_perm(SM_E_NOTFOUND);
}

/*
**  QM_SC_CONF -- find smtpc configuration data for server
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		ipaddr -- IP address of server
**		dadb_entry -- DA DB entry
**
**	Returns:
**		usual sm_error code
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
qm_sc_conf(qmgr_ctx_P qmgr_ctx, sm_ipaddr_P ipaddr, dadb_entry_P dadb_entry)
{
#if MTA_USE_TLS
	sm_ret_T ret;

	dadb_entry->dadbe_se_maprescnf = sm_err_perm(SM_E_NOTFOUND);
	if (NULL == qmgr_ctx->qmgr_conf_map ||
		(!QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_SC_LKP_SE_CONF) &&
		 !SM_IS_FLAG(qmgr_ctx->qmgr_daflags, DA_CFL_SE_FEAT)))
	{
		QM_LEV_DPRINTFC(QDC_Q2C, 9, (QM_DEBFP, "func=%s, map=%p, flags=%#x\n", SMFCT, qmgr_ctx->qmgr_conf_map, qmgr_ctx->qmgr_cnf.q_cnf_flags));
		return SM_SUCCESS;
	}

	if (NULL == dadb_entry->dadbe_se_conf)
		dadb_entry->dadbe_se_conf = sm_str_new(NULL, 256, 1024); /*check size*/
	if (NULL == dadb_entry->dadbe_lhs)
		dadb_entry->dadbe_lhs = sm_str_new(NULL, 10, 16); /* check size */
	if (NULL == dadb_entry->dadbe_tag)
		dadb_entry->dadbe_tag = sm_str_new(NULL, 32, 64); /* check size */
	if (NULL == dadb_entry->dadbe_se_conf ||
	    NULL == dadb_entry->dadbe_lhs ||
	    NULL == dadb_entry->dadbe_tag)
	{
		return sm_err_temp(ENOMEM);
	}

	sm_str_clr(dadb_entry->dadbe_se_conf);
	sm_str_clr(dadb_entry->dadbe_lhs);
	sm_str_clr(dadb_entry->dadbe_tag);

#define SC_SE_TAG "smtpc_session_conf:"
	ret = sm_err_temp(SM_E_OVFLW_NS);
	if (sm_str_scat(dadb_entry->dadbe_tag, SC_SE_TAG) == SM_SUCCESS
	    && sm_ipaddr2str(ipaddr, dadb_entry->dadbe_lhs) == SM_SUCCESS
	    && (ret = sm_map_lookup_ip(qmgr_ctx->qmgr_conf_map,
	                dadb_entry->dadbe_lhs, dadb_entry->dadbe_tag,
	                SMMAP_LFL_SUBNETS|SMMAP_LFL_TAG,
	                dadb_entry->dadbe_se_conf)) == SM_SUCCESS
	    && sm_str_getlen(dadb_entry->dadbe_se_conf) > 0)
	{
QM_LEV_DPRINTFC(QDC_Q2C, 4, (QM_DEBFP, "func=%s, ip=%I, %slookup=%r, rhs=%S\n", SMFCT, ipaddr, SC_SE_TAG, ret, dadb_entry->dadbe_se_conf));
		dadb_entry->dadbe_se_maprescnf = ret;
	}

	return ret;
#else /* MTA_USE_TLS */
	return SM_SUCCESS;
#endif /* MTA_USE_TLS */
}


/*
**  QM_TO_SC_TASK -- create one session with one task for SMTPC
**	XXX HACK just one session with one transaction with one (or more)
**		recipients (see qm_to_sc_add_rcpt())
**
**	Parameters:
**		qsc_ctx -- QMGR SMTPC context
**		se_reuse -- reuse an open session?
**		one_ta -- only one transaction? (close session after this TA?)
**		aq_ta -- AQ transaction
**		aq_rcpt -- AQ recipient
**		rcbe -- RCB entry
**		dadb_entry -- DA DB entry
**
**	Returns:
**		usual sm_error code; see sm_rcb_putv()
**
**	Side Effects: does not clean up rcb in case of an error.
**
**	Last code review: 2005-04-04 21:03:45
**	Last code change: 2006-02-21 05:55:16
*/

static sm_ret_T
qm_to_sc_task(qsc_ctx_P qsc_ctx, bool se_reuse, bool one_ta, aq_ta_P aq_ta, aq_rcpt_P aq_rcpt, sm_rcbe_P rcbe, dadb_entry_P dadb_entry, time_T time_now)
{
	sm_rcb_P rcb;
	sm_ret_T ret;
	uint32_t cdb_rt, ta_flags, se_flags, tmo;
	uint incr_tmo;
	bool dsn;
#if MTA_USE_TLS
	bool sessconf;
# if FFR_CERT_PINNING
	bool prevcert;
# endif
#endif
	extern sm_str_P NullSender;
	sm_str_P sender, verp;

	SM_ASSERT(!aq_rcpt_has_owner(aq_rcpt) ||
		(aq_rcpt->aqr_owner_idx > 0 &&
		 aq_rcpt->aqr_owner_idx <= aq_ta->aqt_owners_n));

	verp = NULL;
	rcb = &rcbe->rcbe_rcb;

	/* treat double bounce differently?? */
	dsn = AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_DSN);
	se_flags = ta_flags = 0;
	cdb_rt = RT_Q2C_CDBID;
	if (dsn &&
	    (QCNF_IS_FLAG(qsc_ctx->qsc_qmgr_ctx, QCNF_FL_B_HDR) ||
	     (AQ_TA_IS_FLAG(aq_ta, AQ_TA_FL_NO_BODY) &&
	      QCNF_IS_FLAG(qsc_ctx->qsc_qmgr_ctx, QCNF_FL_B_BODY_HDR))))
	{
		cdb_rt = RT_Q2C_CDBID_HDR;
		ta_flags |= DA_FL_HDR_ONLY;
	}
	if (dsn && QCNF_IS_FLAG(qsc_ctx->qsc_qmgr_ctx, QCNF_FL_B_DSN_RFC))
		ta_flags |= DA_FL_DSN_RFC;
	else if (dsn && QCNF_IS_FLAG(qsc_ctx->qsc_qmgr_ctx, QCNF_FL_B_DSN_MIME))
		ta_flags |= DA_FL_DSN_MIME;
	if (!one_ta)
		se_flags |= DA_SE_FL_KEEP;
	if (se_reuse)
		se_flags |= DA_SE_FL_REUSE_EX;	/* not used anywhere??? */

	if (dsn) {
		sender = NullSender;
	}
	else if (AQR_IS_FLAG(aq_rcpt, AQR_FL_HAS_VERP) && aq_rcpt_has_owner(aq_rcpt))
	{
		size_t len;
		sm_str_P owner;

		owner = aq_ta->aqt_owners_pa[aq_rcpt->aqr_owner_idx - 1];
		len = sm_str_getlen(owner) + sm_str_getlen(aq_rcpt->aqr_pa) + 1;
		verp = sm_str_new(NULL, len, len + 4);
		if (NULL == verp) {
			ret = sm_err_temp(ENOMEM);
			goto error;
		}
		ret = sm_verpify(owner, aq_rcpt->aqr_pa, NULL,
			qsc_ctx->qsc_qmgr_ctx->qmgr_cnf.q_cnf_verp1,
			qsc_ctx->qsc_qmgr_ctx->qmgr_cnf.q_cnf_verp2,
			qsc_ctx->qsc_qmgr_ctx->qmgr_cnf.q_cnf_verp3,
			verp);
		if (sm_is_err(ret)) goto error;
		sender = verp;
	}
	else if (AQ_TA_IS_FLAG(aq_ta, AQ_TA_FL_VERP)) {
		size_t len;

		len = sm_str_getlen(aq_ta->aqt_mail->aqm_pa)
				+ sm_str_getlen(aq_rcpt->aqr_pa) + 1;
		verp = sm_str_new(NULL, len, len + 4);
		if (NULL == verp) {
			ret = sm_err_temp(ENOMEM);
			goto error;
		}
		ret = sm_verpify(aq_ta->aqt_mail->aqm_pa,
			QCNF_IS_FLAG(qsc_ctx->qsc_qmgr_ctx, QCNF_FL_VERP_RCPT)
				? aq_rcpt->aqr_pa : NULL, NULL,
			qsc_ctx->qsc_qmgr_ctx->qmgr_cnf.q_cnf_verp1,
			qsc_ctx->qsc_qmgr_ctx->qmgr_cnf.q_cnf_verp2,
			qsc_ctx->qsc_qmgr_ctx->qmgr_cnf.q_cnf_verp3,
			verp);
		if (sm_is_err(ret)) goto error;
		sender = verp;
	}
	else if (aq_rcpt_has_owner(aq_rcpt))
		sender = aq_ta->aqt_owners_pa[aq_rcpt->aqr_owner_idx - 1];
	else
		sender = aq_ta->aqt_mail->aqm_pa;

	/* increase timeout if there was any (temp) error before */
#define INCR_TMO_BASE	100
#define INCR_TMO_ADD	50	/* CONF? */
	incr_tmo = INCR_TMO_BASE;
	if (aq_rcpt->aqr_err_fl != 0 && aq_rcpt->aqr_tries > 0)
		incr_tmo += INCR_TMO_ADD * aq_rcpt->aqr_tries;
QM_LEV_DPRINTFC(QDC_SCHED, 2, (QM_DEBFP, "func=qm_to_sc_task, ss_ta=%s, da_ta=%s, stat=%d, err_fl=%x, incr_tmo=%d\n", aq_ta->aqt_ss_ta_id, dadb_entry->dadbe_da_ta_id, aq_rcpt->aqr_status, aq_rcpt->aqr_err_fl, incr_tmo));

	(void) qm_rcpt_da_expire(qsc_ctx->qsc_qmgr_ctx, aq_rcpt, time_now,
			&aq_rcpt->aqr_expire);
	tmo = aq_rcpt->aqr_expire - time_now;

#if MTA_USE_TLS
	sessconf = !se_reuse && sm_is_success(dadb_entry->dadbe_se_maprescnf) &&
	    dadb_entry->dadbe_se_conf != NULL &&
	    sm_str_getlen(dadb_entry->dadbe_se_conf) > 0;
# if FFR_CERT_PINNING
	prevcert = !se_reuse && dadb_entry->dadbe_certfpinfo.tlscp_cert_mdlen > 0 && time_now < dadb_entry->dadbe_certfpinfo.tlscp_notafter_t;
#  if 0
	if (prevcert) {
		QM_LEV_DPRINTFC(QDC_SCHED, 2, (QM_DEBFP,
			"sev=DBG, func=%s, ss_ta=%s, cert_fp=%0.*hhy\n"
			, SMFCT
			, aq_ta->aqt_ss_ta_id
			, dadb_entry->dadbe_certfpinfo.tlscp_cert_mdlen
			, dadb_entry->dadbe_certfpinfo.tlscp_cert_md
			));
	}
#  endif /* 0 debug printing */
		QM_LEV_DPRINTFC(QDC_SCHED, 2, (QM_DEBFP,
			"sev=DBG, func=%s, ss_ta=%s, hs_stat=%m, hs_tricks=%d\n"
			, SMFCT
			, aq_ta->aqt_ss_ta_id
			, dadb_entry->dadbe_certfpinfo.tlscp_hs_stat
			, dadb_entry->dadbe_certfpinfo.tlscp_hs_tricks
			));
# endif
#endif

	ret = sm_rcb_putv(rcb, RCB_PUTV_FIRST,
		SM_RCBV_INT, RT_PROT_VER, PROT_VER_RT,
		SM_RCBV_INT, RT_Q2C_ID, qsc_ctx->qsc_id,
#ifdef RT_Q2C_DCID
		SM_RCBV_INT, RT_Q2C_DCID, 0, /* delivery class, cur. unused */
#endif
		SM_RCBV_INT, RT_Q2C_SE_FLAGS, se_flags,
		SM_RCBV_BUF,
			se_reuse ? (one_ta ? RT_Q2C_SEID_1TA  : RT_Q2C_SEID)
				 : (one_ta ? RT_Q2C_NSEID_1TA : RT_Q2C_NSEID),
			dadb_entry->dadbe_da_se_id, SMTP_STID_SIZE,
		SM_RCBV_INT,
			(0 == aq_rcpt->aqr_port) ? RT_NOSEND : RT_Q2C_SRVPORT,
			(uint32_t) aq_rcpt->aqr_port,
		SM_RCBV_INT,
			(DA_IDX_ESMTP == aq_rcpt->aqr_da_idx) ? RT_NOSEND : RT_Q2C_DA_IDX,
			aq_rcpt->aqr_da_idx,

		SM_RCBV_INT, (INCR_TMO_BASE == incr_tmo) ? RT_NOSEND : RT_Q2C_TMO_M, incr_tmo,
		SM_RCBV_INT, RT_Q2C_EXP, tmo,
		SM_RCBV_INT, RT_Q2C_ERRFL, aq_rcpt->aqr_err_fl,

		/* HACK */
		SM_RCBV_IPADDR, RT_Q2C_SRVIP,
			&(AQR_CUR_DEST(aq_rcpt)),

#if MTA_USE_TLS
		SM_RCBV_INT, sessconf ? RT_Q2C_MAP_RES_CNF_SRV : RT_NOSEND,
			dadb_entry->dadbe_se_maprescnf,
		SM_RCBV_STR, sessconf ? RT_Q2C_RHS_CNF_SRV : RT_NOSEND,
			dadb_entry->dadbe_se_conf,
# if FFR_CERT_PINNING
		SM_RCBV_INT, dadb_entry->dadbe_certfpinfo.tlscp_hs_stat != 0 ? RT_Q2C_HS_STAT : RT_NOSEND, dadb_entry->dadbe_certfpinfo.tlscp_hs_stat,
		SM_RCBV_INT, dadb_entry->dadbe_certfpinfo.tlscp_hs_stat != 0 ? RT_Q2C_HS_TRICKS : RT_NOSEND, dadb_entry->dadbe_certfpinfo.tlscp_hs_tricks,
		SM_RCBV_INT, prevcert ? RT_Q2C_CERT_MDLEN : RT_NOSEND, dadb_entry->dadbe_certfpinfo.tlscp_cert_mdlen,
		SM_RCBV_BUF, prevcert ? RT_Q2C_CERT_MD : RT_NOSEND, dadb_entry->dadbe_certfpinfo.tlscp_cert_md, sizeof(dadb_entry->dadbe_certfpinfo.tlscp_cert_md),
# endif
#endif

		SM_RCBV_BUF,
			AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_DLY)
			? RT_Q2C_NTAIDD : AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_BNC)
			? RT_Q2C_NTAIDB : AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_DBNC)
			? RT_Q2C_NTAIDDB : RT_Q2C_NTAID,
			dadb_entry->dadbe_da_ta_id, SMTP_STID_SIZE,
		SM_RCBV_BUF, RT_Q2C_SSTAID, aq_ta->aqt_ss_ta_id, SMTP_STID_SIZE,
		SM_RCBV_OFF, RT_Q2C_SIZE_B, aq_ta->aqt_msg_sz_b,
		SM_RCBV_STR, RT_Q2C_MAIL, sender,
		SM_RCBV_CSTR, cdb_rt, aq_ta->aqt_cdb_id,
		SM_RCBV_INT, RT_Q2C_TA_FLAGS, ta_flags,
		SM_RCBV_END);
	if (sm_is_success(ret))
		ret = sm_hdrmodl_wr(aq_ta->aqt_hdrmodhd, rcb, RT_Q2C_HM_T_P, RT_Q2C_HM_HDR);

	if (sm_is_success(ret))
		ret = sm_rcb_putv(rcb, RCB_PUTV_NONE,
			SM_RCBV_INT, RT_Q2C_RCPT_IDX, aq_rcpt->aqr_idx,
			SM_RCBV_STR, RT_Q2C_RCPT, aq_rcpt->aqr_pa,
			SM_RCBV_END);

#if MTA_USE_TLS
	if (sm_is_success(ret) &&
	    sm_is_success(aq_rcpt->aqr_maprescnf) &&
	    aq_rcpt->aqr_conf != NULL)
	{
		ret = sm_rcb_putv(rcb, RCB_PUTV_NONE,
			SM_RCBV_INT, RT_Q2C_MAP_RES_CNF_RCPT, aq_rcpt->aqr_maprescnf,
			SM_RCBV_STR, RT_Q2C_RHS_CNF_RCPT, aq_rcpt->aqr_conf,
			SM_RCBV_END);
	}
#endif /* MTA_USE_TLS */

	QM_LEV_DPRINTFC(QDC_SCHED, 1, (QM_DEBFP, "func=qm_to_sc_task, ss_ta=%s, cdb=%s, da_ta=%s, ta_flags=%#x, dsn=%d, ret=%r\n",
		aq_ta->aqt_ss_ta_id, sm_cstr_data(aq_ta->aqt_cdb_id),
		dadb_entry->dadbe_da_ta_id, aq_ta->aqt_flags, dsn, ret));
	if (sm_is_err(ret)) goto error;
	if (dsn) {
		ret = sm_rcb_putv(rcb, RCB_PUTV_NONE,
			SM_RCBV_STR, RT_Q2C_B_MSG, aq_rcpt->aqr_dsn_msg,
			SM_RCBV_END);
		if (sm_is_err(ret)) goto error;
		AQ_TA_SET_FLAG(aq_ta, AQ_TA_FL_NO_BODY);

		if (QCNF_IS_FLAG(qsc_ctx->qsc_qmgr_ctx, QCNF_FL_B_DSN_RFC) &&
			!SM_IS_EMPTY_STR(aq_rcpt->aqr_dsn_rfc))
		{
			ret = sm_rcb_putv(rcb, RCB_PUTV_NONE,
				SM_RCBV_STR, RT_Q2C_R_MSG, aq_rcpt->aqr_dsn_rfc,
				SM_RCBV_END);
			if (sm_is_err(ret)) goto error;
		}
	}
	SM_STR_FREE(verp);
	return ret;

  error:
	SM_STR_FREE(verp);
	QM_LEV_DPRINTFC(QDC_SCHED, 0, (QM_DEBFP, "sev=ERROR, func=qm_to_sc_task, ret=%r\n", ret));
	return ret;
}

#if QMGR_DEBUG > 1
void
aq_rcpt_print(aq_rcpt_P aq_rcpt)
{
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aq_rcpt=%p\n", aq_rcpt));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "ss_ta=%s\n", aq_rcpt->aqr_ss_ta_id));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "rcpt=%S\n", aq_rcpt->aqr_pa));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_idx=%u\n", aq_rcpt->aqr_idx));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_tries=%u\n", aq_rcpt->aqr_tries));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "srv_ip=%I\n", &aq_rcpt->aqr_addrs == NULL ? NULL : &aq_rcpt->aqr_addrs[0].aqra_ip));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_da_idx=%u\n", aq_rcpt->aqr_da_idx));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_st_time=%ld\n", aq_rcpt->aqr_st_time));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_last_try=%ld\n", aq_rcpt->aqr_last_try));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_next_try=%ld\n", aq_rcpt->aqr_next_try));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_status=%d\n", aq_rcpt->aqr_status));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqr_flags=%#x\n", aq_rcpt->aqr_flags));
	QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "aqr_ss_link.sm_rg_succ=%p\n", AQR_SS_SUCC(aq_rcpt)));
	QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "aqr_ss_link.sm_rg_pred=%p\n", AQR_SS_PRED(aq_rcpt)));
	QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "aqr_da_link.sm_rg_succ=%p\n", AQR_DA_SUCC(aq_rcpt)));
	QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "aqr_da_link.sm_rg_pred=%p\n", AQR_DA_PRED(aq_rcpt)));
}

void
aq_ta_print(aq_ta_P aq_ta)
{
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aq_ta=%p\n", aq_ta));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "ss_ta=%s\n", aq_ta->aqt_ss_ta_id));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "mail=%S\n", aq_ta->aqt_mail->aqm_pa));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "time=%ld\n", (long) aq_ta->aqt_st_time));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "cdb=%s\n", sm_cstr_data(aq_ta->aqt_cdb_id)));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "rcpts=%u\n", aq_ta->aqt_rcpts_inaq));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "rcpts_tot=%u\n", aq_ta->aqt_rcpts_tot));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "rcpts_left=%u\n", aq_ta->aqt_rcpts_left));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "rcpts_temp=%u\n", aq_ta->aqt_rcpts_temp));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "rcpts_perm=%u\n", aq_ta->aqt_rcpts_perm));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "state=%d\n", aq_ta->aqt_state));
	QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "aqt_rcpts_ar=%d\n", aq_ta->aqt_rcpts_ar));
}
#endif /* QMGR_DEBUG > 1 */

/*
**  QM_TO_SC_ADD_RCPT -- add another recipient to a transaction for SMTPC
**
**	Parameters:
**		qsc_ctx -- QMGR SMTPC context (currently unused)
**		aq_rcpt -- AQ recipient
**		rcbe -- RCB entry
**
**	Returns:
**		usual sm_error code; see sm_rcb_putv()
**
**	Side Effects: does not clean up rcb in case of an error.
**
**	Last code review: 2005-04-04 21:05:04
**	Last code change:
*/

static sm_ret_T
qm_to_sc_add_rcpt(qsc_ctx_P qsc_ctx, aq_rcpt_P aq_rcpt, sm_rcbe_P rcbe)
{
	sm_rcb_P rcb;
	sm_ret_T ret;

	rcb = &rcbe->rcbe_rcb;
	ret = sm_rcb_putv(rcb, RCB_PUTV_NONE,
		SM_RCBV_INT, RT_Q2C_RCPT_IDX, aq_rcpt->aqr_idx,
		SM_RCBV_STR, RT_Q2C_RCPT, aq_rcpt->aqr_pa,
		SM_RCBV_END);

#if MTA_USE_TLS
	if (sm_is_success(ret) &&
	    sm_is_success(aq_rcpt->aqr_maprescnf) &&
		aq_rcpt->aqr_conf != NULL)
	{
		ret = sm_rcb_putv(rcb, RCB_PUTV_NONE,
			SM_RCBV_INT, RT_Q2C_MAP_RES_CNF_RCPT, aq_rcpt->aqr_maprescnf,
			SM_RCBV_STR, RT_Q2C_RHS_CNF_RCPT, aq_rcpt->aqr_conf,
			SM_RCBV_END);
	}
#endif /* MTA_USE_TLS */
	if (sm_is_err(ret)) goto error;
	return ret;

  error:
	QM_LEV_DPRINTFC(QDC_SCHED, 0, (QM_DEBFP, "sev=ERROR, func=qm_to_sc_add_rcpt, ret=%r\n", ret));
	return ret;
}

/*
**  QMGR_FIND_DA -- Find an available DA
**	NOTE: this requires that all DAs can do all kind of deliveries!
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		pqsc_bit -- (pointer to) QSC bit (output)
**		pqsc_ctx -- (pointer to) QSC context (output)
**			if this points to a valid qsc_ctx: don't return the same one.
**		pda_avail -- (pointer to) how many entries are free? (output)
**		pda_idle -- (pointer to) how many entries are idle? (output)
**
**	Returns:
**		usual sm_error code; SM_E_NO_DA, (un)lock
**
**	Locking: locks qmgr_ctx
**
**	Last code review: 2005-04-04 22:07:13
**	Last code change: 2006-01-29 05:03:32
*/

static sm_ret_T
qmgr_find_da(qmgr_ctx_P qmgr_ctx, uint32_t *pqsc_bit, qsc_ctx_P *pqsc_ctx, uint *pda_avail, uint *pda_idle)
{
	int i, r;
	uint32_t j;
	uint da_avail, da_idle;
	sm_ret_T ret;
	qsc_ctx_P qsc_ctx;

	SM_REQUIRE(pqsc_ctx != NULL);
	SM_REQUIRE(pqsc_bit != NULL);
	*pqsc_bit = 0;
	r = pthread_mutex_lock(&qmgr_ctx->qmgr_mutex);
	SM_LOCK_OK(r);
	if (r != 0) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SMTPS, QM_LMOD_FROM_SMTPS, SM_LOG_CRIT, 4,
			"sev=CRIT, func=qmgr_find_da, lock=%d\n", r);
		return sm_error_temp(SM_EM_Q_SCHED, r);
	}
	ret = sm_error_warn(SM_EM_Q_SCHED, SM_E_NO_DA);	/* XXX */
	for (j = 1, i = 0; i < QM_N_SC_GLI(qmgr_ctx); i++, j *= 2) {
		if ((qmgr_ctx->qmgr_sc_li.qm_gli_used & j) != 0
		    && (qsc_ctx = qmgr_li_sc(qmgr_ctx, i)) != NULL
		    && qsc_ctx->qsc_com.rcbcom_tsk != NULL
		    && QSC_IS_RUNNING(qsc_ctx)
		    && qsc_ctx != *pqsc_ctx
		   )
		{
			/* check whether the DA has free capacity */
			(void) dadb_entry_avail(qsc_ctx->qsc_dadb_ctx, &da_avail, &da_idle, THR_LOCK_UNLOCK);
			if (da_avail > 0 || da_idle > 0) {
				if (pda_avail != NULL)
					*pda_avail = da_avail;
				if (pda_idle != NULL)
					*pda_idle = da_idle;

				*pqsc_ctx = qsc_ctx;
				*pqsc_bit = j;
				ret = SM_SUCCESS;
				break;
			}
		}
	}
	r = pthread_mutex_unlock(&qmgr_ctx->qmgr_mutex);
	SM_ASSERT(0 == r);
	SM_ASSERT(ret != SM_SUCCESS || *pqsc_ctx != NULL);
	if (r != 0 && sm_is_success(ret))
		ret = sm_error_perm(SM_EM_Q_SCHED, r);
	return ret;
}

/*
**  SCHED_MV2NXTMX -- Maybe move rcpt to next MX
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_ctx -- AQ context
**		aq_rcpt -- AQ recipient
**		time_now -- current time
**
**	Returns:
**		usual sm_error code; ENOMEM, SM_E_UNEXPECTED et.al.
**
**	Locking: aq_ctx must be locked
**
**	Side Effects: may remove aq_rcpt from RDQ without adding it
**		to next one due to resource problem
**	Note: fix this! try to add first to new RDQ and only if that succeeds,
**		remove it from the old one.
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
sched_mv2nxtmx(qmgr_ctx_P qmgr_ctx, aq_ctx_P aq_ctx, aq_rcpt_P aq_rcpt, time_T time_now)
{
	sm_ret_T ret;
	uint addr_cur;

	addr_cur = aq_rcpt->aqr_addr_cur;
	if (!(AQR_MORE_ADDR(aq_rcpt)
#if FFR_FBMX
		/* stop at FallbackMX */
	    && aq_rcpt->aqr_addrs[addr_cur].aqra_pref < FBMX_PREF
#endif
#if 0
	    aq_rcpt->aqr_addrs[addr_cur].aqra_pref ==
		aq_rcpt->aqr_addrs[addr_cur + 1].aqra_pref
#endif
	    && aq_rcpt->aqr_addrs[addr_cur].aqra_expt >= time_now))
	{
		return SM_NOTDONE;
	}

	ret = aq_rdq_rm(aq_ctx, aq_rcpt, THR_NO_LOCK, qmgr_ctx->qmgr_lctx);
	if (sm_is_err(ret)) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
			"sev=ERROR, func=sched_mv2nxtmx, aq_rdq_rm=%m, flags=%#x",
			ret, aq_rcpt->aqr_flags);
		return ret;
	}
	aq_rcpt->aqr_addr_cur++;
	ret = aq_rdq_add(aq_ctx, aq_rcpt, NULL /* &aqrdq_flags */, THR_NO_LOCK);
	if (sm_is_err(ret)) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
			"sev=ERROR, func=sched_mv2nxtmx, aq_rdq_add=%m, flags=%#x",
			ret, aq_rcpt->aqr_flags);
		return ret;
	}
	return SM_SUCCESS;
}

/*
**  SCHED2LONG -- AQ rcpt takes too long to schedule
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_ctx -- AQ context
**		aq_rcpt -- AQ recipient
**		time_now -- current time
**		iserr -- is this an error (or a deliberated action)?
**
**	Returns:
**		usual sm_error code
**
**	Locking: aq_ctx must be locked
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
sched2long(qmgr_ctx_P qmgr_ctx, aq_ctx_P aq_ctx, aq_rcpt_P aq_rcpt, time_T time_now, bool iserr)
{
	sm_ret_T ret;

	ret = SM_SUCCESS;
	sm_log_entry(qmgr_ctx->qmgr_lctx,
		QM_LCAT_SCHED, QM_LMOD_SCHED,
		iserr ? SM_LOG_WARN : SM_LOG_INFO, iserr ? 1 : 9,
		SLT_SEV, iserr ? "WARN" : "INFO",
		SLT_FUNC, "sched2long",
		SLT_SS_TA, aq_rcpt->aqr_ss_ta_id,
		SLT_RCPT_IDX, aq_rcpt->aqr_idx,
		SLT_NOW_ENTERED, (long) (time_now - aq_rcpt->aqr_entered),
		SLT_STATUS, "timeout_in_scheduler",
		SLT_END);

	/*
	**  And now?
	**  What's the simplest way to remove this from AQ?
	**  Move it from rcpt dest queue to wait queue and
	**  "cheat" on the timeout ("immediate")?
	*/

	++aq_ctx->aq_t_da;
	AQR_DA_INIT(aq_rcpt);

	/* Was: rdq from todo to busy? check cleanup code (and how rdq is used) */
	ret = aq_rdq_rm(aq_ctx, aq_rcpt, THR_NO_LOCK, qmgr_ctx->qmgr_lctx);
	if (sm_is_err(ret)) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
			"sev=ERROR, func=sched2long, aq_rdq_rm=%m, flags=%#x",
			ret, aq_rcpt->aqr_flags);
		return ret;
	}

	ret = aq_waitq_add(aq_ctx, aq_rcpt, time_now, AQWQ_SCHED, THR_NO_LOCK);
	if (sm_is_err(ret)) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
			"sev=ERROR, func=sched2long, aq_waitq_add=%m",
			ret);
		return ret;
	}
	ret = qmgr_set_aq_cleanup(qmgr_ctx->qmgr_cleanup_ctx, time_now, true);
	if (sm_is_err(ret)) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
			"sev=ERROR, func=sched2long, qmgr_set_aq_cleanup=%m",
			ret);
		return ret;
	}
	return ret;
}

/*
**  SCHED_RCPTMV -- Move RCPT: to defedb or to next MX
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_ctx -- AQ context
**		aq_rcpt -- AQ recipient
**		aqusage -- AQ usage (%)
**		time_now -- current time
**
**	Returns:
**		usual sm_error code; ENOMEM, SM_E_UNEXPECTED et.al.
**
**	Locking: aq_ctx must be locked
**
**	Last code review:
**	Last code change:
*/

static sm_ret_T
sched_rcptmv(qmgr_ctx_P qmgr_ctx, aq_ctx_P aq_ctx, aq_rcpt_P aq_rcpt, int aqusage, time_T time_now)
{
	sm_ret_T ret;

	/* AQ too full? */
	if (aqusage > 90)
		ret = sched2long(qmgr_ctx, aq_ctx, aq_rcpt, time_now, false);
	else
		ret = sched_mv2nxtmx(qmgr_ctx, aq_ctx, aq_rcpt, time_now);
	return ret;
}


/*
**  SAMEDOMAINOK -- Are all recipients for the same domain ready for scheduling?
**
**	Parameters:
**		aq_rcpt -- AQ recipient
**		defaultdomain -- domain to use if recipient doesn't have one
**
**	Returns:
**		true: all ok
**		false: at least one is not yet ready
**
**	Locking: aq_ctx must be locked
**
**	Last code review: 2005-04-04 23:49:13
**	Last code change: 2006-06-11 04:14:31
*/

static bool
samedomainok(aq_rcpt_P aq_rcpt, sm_str_P defaultdomain)
{
	aq_rcpt_P aq_rcpt_nxt;
	bool ok;

	/*
	**  Check whether all recipients with the same domain
	**  are ready for scheduling.
	**  Notes:
	**  1. it is necessary to look in both directions.
	**  this assumes that recipients for the same domain
	**  are treated "similar" by smar; it is not required
	**  that they actually have the same destination,
	**  but if it takes much longer for some of the
	**  recipients to determine the resolved addresses
	**  than for others (even with the same domain! i.e.,
	**  routing based on other criteria), then this may
	**  add unnecessary delays.
	**  2. this function requires that the rcpts are sorted by domain.
	*/

	ok = true;
	for (aq_rcpt_nxt = AQR_SS_SUCC(aq_rcpt);
	     ok && aq_rcpt_nxt != aq_rcpt &&
	     aq_rcpt_eq_domain(aq_rcpt, aq_rcpt_nxt, defaultdomain) == SM_SUCCESS;
	     aq_rcpt_nxt = AQR_SS_SUCC(aq_rcpt_nxt))
	{
		ok = AQR_IS_FLAG(aq_rcpt_nxt, AQR_FL_RDY4DLVRY);
	}
	for (aq_rcpt_nxt = AQR_SS_PRED(aq_rcpt);
	     ok && aq_rcpt_nxt != aq_rcpt &&
	     aq_rcpt_eq_domain(aq_rcpt, aq_rcpt_nxt, defaultdomain) == SM_SUCCESS;
	     aq_rcpt_nxt = AQR_SS_PRED(aq_rcpt_nxt))
	{
		ok = AQR_IS_FLAG(aq_rcpt_nxt, AQR_FL_RDY4DLVRY);
	}
	return ok;
}

/*
**  QMGR_RPCTS_TA -- return max number of recipients for this transaction
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_rcpt -- AQ recipient
**
**	Returns:
**		max number of recipients for this transaction
*/

static uint
qmgr_rpcts_ta(qmgr_ctx_P qmgr_ctx, aq_rcpt_P aq_rcpt)
{
	SM_IS_QMGR_CTX(qmgr_ctx);
	SM_IS_AQ_RCPT(aq_rcpt);

	if (IS_LMTP_ADDR(AQR_CUR_DEST(aq_rcpt)))
		return qmgr_ctx->qmgr_cnf.q_cnf_lmtp_rcpts_ta;
	return qmgr_ctx->qmgr_cnf.q_cnf_smtp_rcpts_ta;
}

/*
**  QMGR_TAS_SE -- return max number of transactions for a session
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_rcpt -- AQ recipient
**		dadb_entry -- DA DB entry
**
**	Returns:
**		max number of transactions for a session
*/

static uint
qmgr_tas_se(qmgr_ctx_P qmgr_ctx, aq_rcpt_P aq_rcpt, dadb_entry_P dadb_entry)
{
	sm_ret_T ret;

	SM_IS_QMGR_CTX(qmgr_ctx);
	SM_IS_AQ_RCPT(aq_rcpt);

	/* get existing limit if possible */
	if (dadb_entry != NULL)
		return dadb_entry->dadbe_max_tas_se;

	if (NULL != qmgr_ctx->qmgr_conf_map &&
		QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_SC_LKP_CL_CONF))
	{
		ret = occ_se_limit(qmgr_ctx, aq_rcpt, qmgr_ctx->qmgr_conf, qmgr_ctx->qmgr_tag);
		if (ret >= 0) {
			sc_cl_P sc_cl;
			uint max_tas_se;

			sc_cl = &(qmgr_ctx->qmgr_cnf.q_cnf_sc_cls.sccls_cl[ret]);
			max_tas_se = sc_cl->sccl_max_tas_se;
			QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "func=qmgr_tas_se, max_tas_se=%u\n", max_tas_se));
			if (max_tas_se > 0)
				return max_tas_se;
		}
	}

	return qmgr_ctx->qmgr_cnf.q_cnf_tas_se;
}

/*
**  QM_SESS_KEEP_OPEN -- is it useful to keep a session open?
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_ta -- AQ transaction
**		aq_rcpt -- AQ recipient
**		dadb_entry -- DA DB entry
**		max_rcpts_ta -- max recipients per transaction
**		max_tas_se -- max transactions per session
**
**	Returns:
**		true iff is it useful to keep a session open
*/

static bool
qm_sess_keep_open(qmgr_ctx_P qmgr_ctx, aq_ta_P aq_ta, aq_rcpt_P aq_rcpt, dadb_entry_P dadb_entry, uint max_rcpts_ta, uint max_tas_se)
{
	uint nrcpts;
	aq_rcpt_P aq_rcpt_nxt;

	if (!QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_SE_REUSE))
		return false;
	aq_rcpt_nxt = AQ_RDQ_NEXT(aq_rcpt);
	if (aq_rcpt_nxt == NULL || aq_rcpt_nxt == aq_rcpt)
{
QM_LEV_DPRINTFC(QDC_SCHED, 4, (QM_DEBFP, "func=qm_sess_keep_open, rcpt_next=NULL, keep_open=0\n"));
		return false;
}

	/* reached limit for number of TAs per session? */
	if (dadb_entry->dadbe_tas_se >= max_tas_se)
{
QM_LEV_DPRINTFC(QDC_SCHED, 4, (QM_DEBFP, "func=qm_sess_keep_open, tas=%d, max_tas=%d, keep_open=0\n", dadb_entry->dadbe_tas_se, max_tas_se));
		return false;
}

	/*
	 *  It is useful to keep a connection open if there are other rcpts
	 *  (for the same destination, that's a given as we look at a RDQ)
	 *  that can not be scheduled in the same transaction.
	 *  By default, we return false.
	 *  It is useful to keep a session open iff
	 *  there is another rcpt that doesn't fit into the same transaction
	 */

	for (nrcpts = 0;
	     aq_rcpt_nxt != NULL && aq_rcpt_nxt != aq_rcpt && nrcpts < max_rcpts_ta;
	     aq_rcpt_nxt = AQ_RDQ_NEXT(aq_rcpt_nxt))
	{
QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "func=qm_sess_keep_open, ss_ta=%s, ss_ta2=%s, same_ta=%d\n", aq_rcpt->aqr_ss_ta_id, aq_rcpt_nxt->aqr_ss_ta_id, SAME_TA(aq_rcpt, aq_rcpt_nxt)));
		if ((AQ_TA_IS_FLAG(aq_ta, AQ_TA_FL_VERP) ||
			!SAME_TA(aq_rcpt, aq_rcpt_nxt))
		   )
		{
QM_LEV_DPRINTFC(QDC_SCHED, 4, (QM_DEBFP, "func=qm_sess_keep_open, ss_ta=%s, ss_ta2=%s, keep_open=1\n", aq_ta->aqt_ss_ta_id, aq_rcpt_nxt->aqr_ss_ta->aqt_ss_ta_id));
			return true;
		}
		++nrcpts;
	}
QM_LEV_DPRINTFC(QDC_SCHED, 4, (QM_DEBFP, "func=qm_sess_keep_open, ss_ta=%s, keep_open=0, nrcpts=%u\n", aq_ta->aqt_ss_ta_id, nrcpts));
	return false;
}

/*
**  QMGR_SCHED_DLVRY -- simple scheduler...
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**		aq_ctx -- AQ context
**		pqsc_bits -- (bitmask) which DAs need to be started (output)
**		pdelay_next_try -- (pointer to) delay until next try (output)
**
**	Returns:
**		usual sm_error code
**
**	Locking: aq_ctx must be locked
**
**	Todo: this is not very sophisticated.
**		split this in multiple functions? it's too long...
**
**	Last code review:
**	Last code change:
*/

/*
**  Delay for scheduler after encountering too many open connections.
**  The scheduler should be woken up after a TA close RCB is received
**  from a DA for which a TA is waiting. This data should be recorded
**  in the DA connection cache.
*/

#define DELAY_TOO_MANY	300
#define SET_DELAY_NEXT_TRY(d, where) do {						\
		if ((d) < *pdelay_next_try || *pdelay_next_try == 0) {	\
			*pdelay_next_try = (d);				\
			QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "func=%s, where=%s, delay=%d\n", SMFCT, (where), (d)));	\
		}							\
	} while (0)

static sm_ret_T
qmgr_sched_dlvry(qmgr_ctx_P qmgr_ctx, aq_ctx_P aq_ctx, uint32_t *pqsc_bits, int *pdelay_next_try)
{
	sm_ret_T ret, res;
	aq_rcpt_P aq_rcpt, aq_rcpt_nxt;
	aq_ta_P aq_ta;
	qsc_ctx_P qsc_ctx;
	sm_rcbe_P rcbe;
	dadb_entry_P dadb_entry;
	aq_rsnd_ctx_P aq_rsnd_ctx;
	uint i, nrcpts, connlimit, conncur, max_rcpts_ta, max_tas_se, da_avail, da_idle;
	int r, aqusage;
	uint32_t qsc_bit;
	bool stopit, se_reuse, se_keep_open, occe_locked;
	time_T time_now;
	sessta_id_T da_ta_id;
	occ_entry_P occ_entry;
	aqrdq_ctx_P aqrdq;
	uint todo_entries, rdq_entries;

	/*
	**  Macro to move rcpt to end of list unless it's there already.
	*/

#define SM_MV_RCPT2END	\
	if (aq_rcpt == AQ_RDQ_LAST(aqrdq->aqrdq_rcpts))	\
		break;						\
	AQ_RDQ_REMOVE(aqrdq->aqrdq_rcpts, aq_rcpt);	\
	AQ_RDQ_INSERT_TAIL(aqrdq->aqrdq_rcpts, aq_rcpt);	\
	continue

#define QM_SCHED_RCBE_FREE			\
	while (rcbe != NULL) {				\
		aq_rsnd_ctx_free(aq_rsnd_ctx);	\
		aq_rsnd_ctx = NULL;		\
		sm_rcbe_free(rcbe);		\
		rcbe = NULL;			\
	}

	SM_IS_QMGR_CTX(qmgr_ctx);
	SM_IS_AQ(aq_ctx);
	SM_REQUIRE(pqsc_bits != NULL);
	SM_REQUIRE(pdelay_next_try != NULL);
	*pqsc_bits = 0;
	da_ta_id[0] = '\0';
	rcbe = NULL;
	aq_rsnd_ctx = NULL;
	occe_locked = stopit = false;
	max_rcpts_ta = INT_MAX / 2;
	max_tas_se = DA_TAS_SE_MAX;

	/*
	**  Do we have a DA? Just one...
	**  Note: this may sleep() while holding AQ mutex, but only
	**  if there was never a DA available before (QMGR_SFL_HAD_DA),
	**  i.e., during startup.
	*/

	qsc_ctx = NULL;
	while (!sm_is_success(ret =
		qmgr_find_da(qmgr_ctx, &qsc_bit, &qsc_ctx, NULL, NULL)))
	{
		if (qmgr_ctx->qmgr_st_time + qmgr_ctx->qmgr_cnf.q_cnf_wait4clt
				<= evthr_time(qmgr_ctx->qmgr_ev_ctx)
		    || QMGR_IS_SFLAG(qmgr_ctx, QMGR_SFL_HAD_DA))
		{
			if (!QMGR_IS_SFLAG(qmgr_ctx, QMGR_SFL_DA)) {
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_WARN, 8,
					"sev=WARN, func=%s, status=no delivery agent available", SMFCT);
				QMGR_SET_SFLAG(qmgr_ctx, QMGR_SFL_DA);
			}
			qmgr_ctx->qmgr_tm_no_da = evthr_time(qmgr_ctx->qmgr_ev_ctx);
			return sm_error_warn(SM_EM_Q_SCHED, SM_E_NO_DA);
		}
		if (qmgr_is_stop(qmgr_ctx))
			return SM_SUCCESS;
		sleep(1);
	}
	if (!QMGR_IS_SFLAG(qmgr_ctx, QMGR_SFL_HAD_DA))
		QMGR_SET_SFLAG(qmgr_ctx, QMGR_SFL_HAD_DA);
	QMGR_CLR_SFLAG(qmgr_ctx, QMGR_SFL_DA);

	SM_IS_QSC_CTX(qsc_ctx);
	QM_LEV_DPRINTFC(QDC_SCHED, 7, (QM_DEBFP, "func=%s, qsc_ctx=%p\n", SMFCT, qsc_ctx));
	SM_IS_EVTHR_TSK(qsc_ctx->qsc_com.rcbcom_tsk);

	/* recalculate in loop? */
	aqusage = aq_usage(aq_ctx, AQ_USAGE_ALL);
	ret = SM_SUCCESS;
	time_now = evthr_time(qmgr_ctx->qmgr_ev_ctx);

	/*
	**  Go through all rcpt destination queues.
	**
	**  Note: currently this code assumes that all recipients can be
	**  delivered by all delivery agents.
	**  Question: what "breaks" if different delivery agents or different
	**  delivery classes need to be selected?
	**  Is is necessary to have rcpt destination queues per delivery class?
	**  Is is necessary to have delivery superclasses, i.e., sets of
	**  delivery classes which are implemented by the same set of DAs
	**  (such that it doesn't make a difference if any of the delivery
	**  classes inside a superclass is selected because they all share
	**  the same DAs)?
	*/

	QM_LEV_DPRINTFC(QDC_SCHED, 6, (QM_DEBFP, "func=%s, rdq_entries=%d\n", SMFCT, aq_ctx->aq_rdq_used));
	for (rdq_entries = aq_ctx->aq_rdq_used; rdq_entries > 0; rdq_entries--) {
		/*
		**  Take the head of the list, and "rotate" it to the end.
		**  If there would be a "rotate" operation, then this wouldn't
		**  be necessary unless a problem occurs in which case the
		**  current entry would become the first element of the list
		**  to achieve some kind of fairness.
		*/

		aqrdq = AQ_RDQS_FIRST(aq_ctx->aq_rdqs);
		AQ_RDQS_REMOVE(aq_ctx->aq_rdqs, aqrdq);
		AQ_RDQS_INSERT_TAIL(aq_ctx->aq_rdqs, aqrdq);
		AQRDQ_CLR_FLAG(aqrdq, AQRDQ_FL_OCEXC);

		todo_entries = aqrdq->aqrdq_entries;
		QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "func=%s, todo_entries=%d\n", SMFCT, todo_entries));

		/*
		**  Note: the loop construct is a problem:
		**  1. aq_rcpt might be removed (most likely)
		**	hence AQ_RDQ_NEXT(aq_rcpt) can't be done inside
		**	the "for" statement.
		**  2. aq_rcpt_nxtl might be removed from the TODO queue too
		**	by the "add more recipients to this transaction" code.
		**  Possible solutions:
		**  1. when removing a recipient from the TODO list, check it
		**  against aq_rcpt_nxtl: if it is the same: go to the next.
		**  2. check whether aq_rcpt_nxtl is still in TODO list,
		**  if not: what then? we can't go to the next entry: this
		**  isn't in the right list...
		**
		**  It seems the loop should be restructured:
		**  all entries in the TODO list are ready for delivery,
		**  hence after each iteration the head of the list should
		**  have been removed from the list, i.e.,
		**  while (!empty) { rcpt = head(todo_list); schedule }
		**
		**  However, even that is incorrect because some recipients
		**  might not be scheduled despite being in the todo list.
		**  See below for two cases.
		**  The hack applied in those cases (move the item to the
		**  end of the list) doesn't work if this happens more than
		**  once in a list. Hence this restriction must be used
		**  unless a better way, e.g., a flag, can be found.
		**
		**  The current solution:
		**  use a for loop restricted by the number of todo entries
		**  always access the first element of the list
		**	rcpts that aren't scheduled must be moved to the end
		**  stop if the list is empty
		**	this can happen if the scheduler adds more rcpts
		**	inside the loop to a transaction
		*/

		/* go through all "todo" rcpts for each rcpt dest queue */
		for (i = 0;
		     i < todo_entries && !stopit && !AQ_RDQ_EMPTY(aqrdq->aqrdq_rcpts);
		     i++)
		{
			da_avail = da_idle = 0;
			aq_rcpt = AQ_RDQ_FIRST(aqrdq->aqrdq_rcpts);
			SM_IS_AQ_RCPT(aq_rcpt);
#if QMGR_TEST
			if (SM_IS_FLAG(qmgr_ctx->qmgr_cnf.q_cnf_tests, QMGR_TEST_DLY_SCHED)
			    && AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_DLY))
			{
				SM_ASSERT(false);	/* abort */
			}
#endif

			aq_ta = aq_rcpt->aqr_ss_ta;
			QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "sev=DBG, func=%s, aq_rcpt=%p, aqr_pa=%S, ss_ta=%s, idx=%u, aqr_flags=%#x, sched=%d, aq_ta=%p\n", SMFCT, aq_rcpt, aq_rcpt->aqr_pa, aq_rcpt->aqr_ss_ta_id, aq_rcpt->aqr_idx, aq_rcpt->aqr_flags, AQR_SCHEDULE(aq_rcpt), aq_ta));
#if QMGR_DEBUG > 1
			aq_rcpt_print(aq_rcpt);
#endif

			SM_IS_AQ_TA(aq_ta);
#if QMGR_DEBUG > 1
			aq_ta_print(aq_ta);
#endif

			SM_ASSERT(!AQR_IS_FLAG(aq_rcpt, AQR_FL_WAITQ_AR));
			SM_ASSERT(!AQR_IS_FLAG(aq_rcpt, AQR_FL_TMOUT|AQR_FL_SCHED|AQR_FL_PERM));
			SM_ASSERT(!AQR_IS_DSNFL(aq_rcpt, AQR_DSNFL_F_HBG));
			SM_ASSERT(AQR_SCHEDULE(aq_rcpt));

			if (AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_DSN))
				QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "sev=DBG, func=%s, aq_rcpt=%p, aq_rcpt_pa=%S, idx=%u, type=bounce, tried=%u, total=%u, now-entered=%ld\n", SMFCT, aq_rcpt, aq_rcpt->aqr_pa, aq_rcpt->aqr_idx, aq_ta->aqt_rcpts_tried, aq_ta->aqt_rcpts_tot, (long) (time_now - aq_rcpt->aqr_entered)));

			/*
			**  Note: the timeout must be significantly larger
			**  than the SMTPC delivery timeout!
			**  Check whether the item itself is too old:
			**  this can cause problems if there are many
			**  destinations that are unavailable (many delivery attempts)
			**  and check whether the time since the last
			**  scheduling attempt is exceeded too.
			**  Should this check the time since the last
			**  "status update" (e.g., scheduled, result from DA
			**  result from AR, others?)
			*/

			if (aq_rcpt->aqr_entered + qmgr_ctx->qmgr_tmo_sched < time_now &&
			    aq_rcpt->aqr_last_try + qmgr_ctx->qmgr_tmo_sched < time_now &&
			    qmgr_ctx->qmgr_tm_no_da + qmgr_ctx->qmgr_tmo_sched < time_now)
			{
				/*
				**  This is taking too long...  something must be wrong.
				*/

				ret = sched2long(qmgr_ctx, aq_ctx, aq_rcpt, time_now, true);
				if (sm_is_err(ret)) {
					stopit = true;
					break;
				}
				continue;
			}

			/*
			**  If this is a (double)bounce then skip it if
			**  - there aren't enough tried recipients (XXX?)
			**  or
			**  - the timeout hasn't occurred yet.
			**
			**  Note: this is one place where a recipient can be in
			**  the todo queue but it is skipped by the scheduler.
			**  See also next if statement.
			*/

			if (AQR_IS_FLAG(aq_rcpt, AQR_FL_IS_DSN) &&
			    !(aq_ta->aqt_rcpts_tot - aq_ta->aqt_rcpts_tried <=
					(aq_ta_has_bounce(aq_ta) ? 1 : 0) +
					(aq_ta_has_dbl_bounce(aq_ta) ? 1 : 0)
			      || aq_rcpt->aqr_entered + qmgr_ctx->qmgr_cnf.q_cnf_t_sched_dsn
				 < time_now)
			   )
			{
				int d;

				d = aq_rcpt->aqr_entered +
				    qmgr_ctx->qmgr_cnf.q_cnf_t_sched_dsn - time_now;
				if (0 == d)
					d = 1;	/* need a minimum delay */
				SET_DELAY_NEXT_TRY(d, "bounce/timeout");

				SM_MV_RCPT2END;
			}

			/*
			**  XXX Hack: need
			**  no more recipients to receive from AR
			**  or
			**  the number of recipients is less than 10
			**  (i.e., for small numbers all recipients must be
			**  resolved before continuing).
			**
			**  This is only a hack anyway, we need a real scheduler...
			**
			**  Note: aqt_rcpts_ar counts also bounce recipients
			**  that have been sent to AR! Should it do that?
			**
			**  Note: this is another place where a recipient can
			**  be in the todo queue but it is skipped by the
			**  scheduler. See also previous if statement.
			*/

			QM_LEV_DPRINTFC(QDC_SCHED, 6, (QM_DEBFP, "sev=DBG, func=%s, aqr_pa=%S, stat=%d, aqt_rcpts_ar=%u, aqt_rcpts_inaq=%u, aq_entries=%u, t_da=%u, aqt_rcpts_arf=%u, skip=%d\n", SMFCT, aq_rcpt->aqr_pa, aq_rcpt->aqr_status, aq_ta->aqt_rcpts_ar, aq_ta->aqt_rcpts_inaq, aq_ctx->aq_entries, aq_ctx->aq_t_da, aq_ta->aqt_rcpts_arf, !((0 == aq_ta->aqt_rcpts_ar) || (aq_ta->aqt_rcpts_ar <= (aq_ta->aqt_rcpts_inaq >> 1) && aq_ta->aqt_rcpts_inaq > 10))));

			/* CONF: timeout waiting for other RCPTs? */
			r = SM_MIN(qmgr_ctx->qmgr_cnf.q_cnf_tmo_ar, 10);
			if (r < 1)
				r = 1;
			if (!((0 == aq_ta->aqt_rcpts_ar) ||
			      (aq_ta->aqt_rcpts_ar <= (aq_ta->aqt_rcpts_inaq >> 1)
				   && aq_ta->aqt_rcpts_inaq > 10
				   && time_now <= aq_rcpt->aqr_entered + r
			      )
			     )
			   )
			{
				r = aq_rcpt->aqr_entered + r + 1 - time_now;
				if (r > 0)
					SET_DELAY_NEXT_TRY(r, "waiting_for_AR_results");
				SM_MV_RCPT2END;
			}

			/*
			**  Check whether all recipients with the same domain
			**  are ready for scheduling.
			*/

			if (!samedomainok(aq_rcpt, qmgr_ctx->qmgr_hostname)) {
				QM_LEV_DPRINTFC(QDC_SCHED, 5, (QM_DEBFP, "sev=DBG, func=%s, aqr_pa=%S, status=skipped_because_other_entry_is_not_ready\n", SMFCT, aq_rcpt->aqr_pa));
				SM_MV_RCPT2END;
			}

			/* check whether current DA can be used */
			ret = dadb_entry_avail(qsc_ctx->qsc_dadb_ctx, &da_avail, &da_idle, THR_LOCK_UNLOCK);
			if (da_avail <= 0 && da_idle <= 0) {
				/* find another DA */
				ret = qmgr_find_da(qmgr_ctx, &qsc_bit, &qsc_ctx, &da_avail, &da_idle);
				if (sm_is_err(ret)) {
					if (sm_error_warn(SM_EM_Q_SCHED, SM_E_NO_DA) == ret)
						qmgr_ctx->qmgr_tm_no_da = time_now;
					QM_LEV_DPRINTFC(QDC_SCHED, 0, (QM_DEBFP, "sev=DBG, func=%s, where=3, qsc_ctx=%p, status=no_free_DA\n", SMFCT, qsc_ctx));
					stopit = true;
					break;
				}
				QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "sev=DBG, func=%s, switch, qsc_ctx=%p\n", SMFCT, qsc_ctx));
			}

			/* add delivery information to SMTPC RCB entry */
			ret = qm_rcbe_new(qmgr_ctx, &rcbe, -1, THR_LOCK_UNLOCK);
			if (sm_is_err(ret)) {
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
					"sev=ERROR, func=%s, qm_rcbe_new=%m", SMFCT, ret);
				stopit = true;
				break;
			}

			/* create callback context for RCB send */
			ret = aq_rsnd_ctx_new(aq_ctx, aq_rcpt, &aq_rsnd_ctx);
			if (sm_is_err(ret) || NULL == aq_rsnd_ctx) {
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
					"sev=ERROR, func=%s, aq_rsnd_ctx_new=%m", SMFCT, ret);
				sm_rcbe_free(rcbe);
				rcbe = NULL;
				stopit = true;
				break;
			}
			sm_rcbe_setcb(rcbe, aq_rcptsent2da, (void *)aq_rsnd_ctx);

			dadb_entry = NULL;
			occ_entry = NULL;
			connlimit = qmgr_ctx->qmgr_cnf.q_cnf_max_conc_conn;
			conncur = 0;

			/* Need to keep occ_entry locked! */
			res = occ_entry_find(qmgr_ctx->qmgr_occ_ctx->occx_ht,
				&AQR_CUR_DEST(aq_rcpt),
				&occ_entry, &qmgr_ctx->qmgr_occ_ctx->occx_mutex,
				THR_LOCK_UNLERR);
			if (SM_SUCCESS == res && occ_entry != NULL) {
				occe_locked = true;
				SM_ASSERT(occ_entry->occe_open_se >= occ_entry->occe_open_ta);
				connlimit = occ_entry->occe_cur_conc;
				conncur = occ_entry->occe_open_se;
				if (conncur >= connlimit && conncur == occ_entry->occe_open_ta)
				{
					AQRDQ_SET_FLAG(aqrdq, AQRDQ_FL_OCEXC);
					OCCE_SET_FLAG(occ_entry, OCCE_FL_SE_WAIT);
					if (QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_SE_REUSE)) {
						/* always? */
						OCCE_SET_FLAG(occ_entry, OCCE_FL_TA_WAIT);
					}

					r = occ_entry->occe_last_upd + occ_entry->occe_timeout - time_now;
					QM_LEV_DPRINTFC(QDC_SCHED, 2, (QM_DEBFP, "sev=DBG, func=%s, ip=%I, occ=limit_exceeded, last_upd=%ld, timeout=%u, now=%ld, r=%d\n", SMFCT,
&AQR_CUR_DEST(aq_rcpt),
(long) occ_entry->occe_last_upd, occ_entry->occe_timeout, (long) time_now, r));
					if (r <= 0)
						r = DELAY_TOO_MANY;
					SET_DELAY_NEXT_TRY(r, "exceeded");
					sm_log_entry(qmgr_ctx->qmgr_lctx,
						QM_LCAT_SCHED, QM_LMOD_SCHED,
						SM_LOG_INFO,
						OCCE_IS_FLAG(occ_entry, OCCE_FL_LOGEXC) ? 14 : 11,
						SLT_SEV, "INFO",
						SLT_FUNC, "qmgr_sched_dlvry",
						SLT_SS_TA, aq_rcpt->aqr_ss_ta_id,
						SLT_RCPT_IDX, aq_rcpt->aqr_idx,
						SLT(SLT_IP, 'I'), &AQR_CUR_DEST(aq_rcpt),
						SLT_OPN_CONN, occ_entry->occe_open_se,
						SLT_CUR_LIMIT, occ_entry->occe_cur_conc,
						SLT_STATUS, "limit_exceeded",
						SLT_TIMEOUT, r,
						SLT_END);
					OCCE_SET_FLAG(occ_entry, OCCE_FL_LOGEXC);
					r = smthread_mutex_unlock(&qmgr_ctx->qmgr_occ_ctx->occx_mutex);
					SM_ASSERT(0 == r);
					if (0 == r)
						occe_locked = false;

					ret = sched_rcptmv(qmgr_ctx, aq_ctx, aq_rcpt, aqusage, time_now);
					QM_SCHED_RCBE_FREE;
					break;
				}
#if OCC_RATE
				else if (occ_entry->occe_max_rate > 0 &&
					occ_rate_exceeded(occ_entry, time_now))
				{
					/*
					 *  calculate when rate is not exceeded
					 *  60: time "window" for rate (rate: n/minute)
					 *  note: occe_rate_first has a granularity of CHTGRAN
					 */

					r = occ_entry->occe_rate_first + 60 - time_now;
					QM_LEV_DPRINTFC(QDC_SCHED, 2, (QM_DEBFP, "sev=DBG, func=%s, occ=rate_exceeded, last_upd=%ld, timeout=%u, now=%ld, r=%d\n", SMFCT,
(long) occ_entry->occe_last_upd, occ_entry->occe_timeout, (long) time_now, r));
					if (r <= 0)
						r = DELAY_TOO_MANY;
					SET_DELAY_NEXT_TRY(r, "rate_exceeded");
					sm_log_entry(qmgr_ctx->qmgr_lctx,
						QM_LCAT_SCHED, QM_LMOD_SCHED,
						SM_LOG_INFO,
						OCCE_IS_FLAG(occ_entry, OCCE_FL_LOGEXC) ? 14 : 11,
						SLT_SEV, "INFO",
						SLT_FUNC, "qmgr_sched_dlvry",
						SLT_SS_TA, aq_rcpt->aqr_ss_ta_id,
						SLT_RCPT_IDX, aq_rcpt->aqr_idx,
						SLT(SLT_IP, 'I'), &AQR_CUR_DEST(aq_rcpt),
						SLT_CUR_RATE, occ_entry->occe_rate,
						SLT_RATE_MAX, occ_entry->occe_max_rate,
						SLT_STATUS, "limit_exceeded",
						SLT_TIMEOUT, r,
						SLT_END);
					OCCE_SET_FLAG(occ_entry, OCCE_FL_LOGEXC);
					r = smthread_mutex_unlock(&qmgr_ctx->qmgr_occ_ctx->occx_mutex);
					SM_ASSERT(0 == r);
					if (0 == r)
						occe_locked = false;

					ret = sched_rcptmv(qmgr_ctx, aq_ctx, aq_rcpt, aqusage, time_now);
					QM_SCHED_RCBE_FREE;
					break;
				}
#endif /* OCC_RATE */
				else {
					OCCE_CLR_FLAG(occ_entry, OCCE_FL_LOGEXC);
				}
			}
/*
**  move into previous if-block: an open session can only be found if
**  occ_entry_find() found an entry.
*/
			if (NULL == dadb_entry
				/* && QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_SE_REUSE) */)
			{
				/*
				**  Find a DA that has a session open to the required IP addr
				*/

				res = dadb_se_find_by_ip(qsc_ctx->qsc_dadb_ctx,
					&AQR_CUR_DEST(aq_rcpt),
					time_now, &dadb_entry);
			}

			sm_log_write(qmgr_ctx->qmgr_lctx,
				QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_INFO, 15,
				"func=%s, ip=%I, open_connections=%d, dadb_entry=%p", SMFCT,
				&AQR_CUR_DEST(aq_rcpt), res, dadb_entry);

			/* restrict number of connections */
			if (dadb_entry == NULL &&
			    ((res > 0 && (uint)res >= connlimit) || conncur >= connlimit))
			{
				/* WHY is occe_locked supposed to be "true" here? */
				SM_ASSERT(occe_locked);
				if (occe_locked) {
					r = smthread_mutex_unlock(&qmgr_ctx->qmgr_occ_ctx->occx_mutex);
					SM_ASSERT(0 == r);
					if (0 == r)
						occe_locked = false;
				}
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_INFO, 13,
					"func=%s, ip=%I, open_connections=%d, status=limit_exceeded", SMFCT,
					&AQR_CUR_DEST(aq_rcpt),
					conncur > 0 ? conncur : res);

					SET_DELAY_NEXT_TRY(DELAY_TOO_MANY, "exceeded2");
					break;
			}

			max_tas_se = qmgr_tas_se(qmgr_ctx, aq_rcpt, dadb_entry);

			se_reuse = res > 0 && dadb_entry != NULL
				&& QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_SE_REUSE)
				&& max_tas_se > dadb_entry->dadbe_tas_se;

			/*
			 *  Note: if the session should not be reused because a
			 *  limit is reached, then the code below must take that
			 *  into account by closing the existing session!
			 *  Currently, this can happen if a limit changes during
			 *  operation. However, to avoid this problem for now,
			 *  we store the limit in the dadb_entry, thus "fixing" it
			 *  for an existing session (it will apply to a new session).
			 */

QM_LEV_DPRINTFC(QDC_SCHED, 1, (QM_DEBFP, "sev=DBG, func=%s, se_reuse=%d, dadb_entry=%p, max_tas_se=%u, tas_se=%u, da_se=%s, ss_ta=%s, rcpt=%@S\n", SMFCT, se_reuse, dadb_entry, max_tas_se
,(dadb_entry != NULL) ? dadb_entry->dadbe_tas_se : 0
,(dadb_entry != NULL) ? dadb_entry->dadbe_da_se_id : "?"
,aq_rcpt->aqr_ss_ta_id, aq_rcpt->aqr_pa
)
);


			if (se_reuse) {
				SM_ASSERT(occe_locked);

				/* let's reuse this ... */
				ret = dadb_sess_reuse(qsc_ctx, qsc_ctx->qsc_dadb_ctx,
					aq_rcpt->aqr_ss_ta_id, dadb_entry, occ_entry, aq_rcpt);

				r = smthread_mutex_unlock(&qmgr_ctx->qmgr_occ_ctx->occx_mutex);
				SM_ASSERT(0 == r);
				if (0 == r)
					occe_locked = false;

				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_INFO, 10,
					"sev=INFO, func=%s, open_connections=%d, ip=%I, dadb_sess_reuse=%m", SMFCT,
					conncur > 0 ? conncur : res,
					&AQR_CUR_DEST(aq_rcpt), ret);

				if (sm_is_err(ret)) {
					sm_log_write(qmgr_ctx->qmgr_lctx,
						QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
						"sev=ERROR, func=%s, dadb_sess_reuse=%m", SMFCT, ret);

					/*
					**  if session can't be reused:
					**  check whether connection limit is exceeded
					*/

					if ((uint)res >= connlimit || conncur >= connlimit) {
						sm_log_write(qmgr_ctx->qmgr_lctx,
							QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_INFO, 11,
							"sev=INFO, func=%s, ip=%I, open_connections=%d, status=session_cannot_be_reused+limit_exceeded", SMFCT,
							&AQR_CUR_DEST(aq_rcpt), res);
						SET_DELAY_NEXT_TRY(DELAY_TOO_MANY, "exceeded3");
						break;
					}
				}
			}
			else if (occe_locked) {
				r = smthread_mutex_unlock(&qmgr_ctx->qmgr_occ_ctx->occx_mutex);
				SM_ASSERT(0 == r);
				if (0 == r)
					occe_locked = false;
			}

			/* can't reuse existing connection? try a new one */
			if ((res == 0 || dadb_entry == NULL || sm_is_err(ret))
			    && da_avail > 0)
			{
				se_reuse = false;

				/* open a DA session */
				ret = dadb_sess_open(qsc_ctx, qsc_ctx->qsc_dadb_ctx,
					aq_rcpt->aqr_ss_ta_id,
					&AQR_CUR_DEST(aq_rcpt),
					aq_rcpt, &dadb_entry);
				if (sm_is_err(ret)) {
					sm_log_write(qmgr_ctx->qmgr_lctx,
						QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
						"sev=ERROR, func=%s, qsc_ctx=%p, dadb_sess_open=%m", SMFCT,
						qsc_ctx, ret);

#if QMGR_DEBUG
					{
						dadb_entry_P dadb_entryh;
						dadb_ctx_P dadb_ctx;
						uint l;

						dadb_ctx = qsc_ctx->qsc_dadb_ctx;
						QM_LEV_DPRINTFC(QDC_SCHED, 0, (QM_DEBFP, "sev=DBG, func=%s, where=4, qsc_ctx=%p, entries_cur=%d, entries_max=%d\n", SMFCT,
qsc_ctx, dadb_ctx->dadb_entries_cur, dadb_ctx->dadb_entries_max));
						for (l = 0; l < dadb_ctx->dadb_entries_max; l++)
						{
							dadb_entryh = (dadb_ctx->dadb_entries)[l];
							QM_LEV_DPRINTFC(QDC_SCHED, 1, (QM_DEBFP, "sev=DBG, func=%s, dadb_entry=%p, flags=%#x\n", SMFCT,
							dadb_entryh, dadb_entryh == NULL ? UINT_MAX : dadb_entryh->dadbe_flags));
						}
					}
#endif /* QMGR_DEBUG */

					/* should this stop the scheduler??? */
					stopit = true;
					break;
				}
				SM_ASSERT(dadb_entry != NULL);
				dadb_entry->dadbe_max_tas_se = max_tas_se;
			}

			if (NULL == dadb_entry) {
				/* no free da for this one: try next rdq */
				QM_SCHED_RCBE_FREE;
				break;
			}

			ret = qm_sc_conf(qmgr_ctx, &AQR_CUR_DEST(aq_rcpt), dadb_entry);
			if (SM_IS_TEMP_ERR(ret)) {
				/* TODO: include se/ta id? */
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_INFO, 9,
					"sev=INFO, func=%s, ip=%I, qm_sc_conf=%r"
					, SMFCT, &AQR_CUR_DEST(aq_rcpt), ret);
				QM_SCHED_RCBE_FREE;
				break;
			}


			/*
			**  Keep the connection open?
			**  Only if it is configured and if there is at least
			**  one other entry in this "ready queue".
			*/

			max_rcpts_ta = qmgr_rpcts_ta(qmgr_ctx, aq_rcpt);
			se_keep_open = qm_sess_keep_open(qmgr_ctx, aq_ta, aq_rcpt,
						dadb_entry, max_rcpts_ta, max_tas_se);

			if (se_keep_open && NULL != occ_entry) {
				/* always? only if limit is reached? */
				OCCE_SET_FLAG(occ_entry, OCCE_FL_TA_WAIT);
			}

			QM_LEV_DPRINTTC(QDC_SCHED, 1, (QM_DEBFP, "sev=DBG, func=%s, session_reuse=%d, keep_open=%d, da_se=%s, da_ta=%s, rcpt=%@S\n", SMFCT,
se_reuse, se_keep_open, dadb_entry->dadbe_da_se_id, dadb_entry->dadbe_da_ta_id, aq_rcpt->aqr_pa), qmgr_ctx->qmgr_ev_ctx->evthr_c_time);
			ret = qm_to_sc_task(qsc_ctx, se_reuse, !se_keep_open,
					aq_ta, aq_rcpt, rcbe, dadb_entry, time_now);
			if (sm_is_err(ret)) {
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
					"sev=ERROR, func=%s, qm_to_sc_task=%m", SMFCT, ret);
				stopit = true;
				break;
			}
			++aq_ctx->aq_t_da;
			SESSTA_COPY(aq_rcpt->aqr_da_ta_id, dadb_entry->dadbe_da_ta_id);
			AQR_DA_INIT(aq_rcpt);

			ret = aq_rdq_rm(aq_ctx, aq_rcpt, THR_NO_LOCK, qmgr_ctx->qmgr_lctx);
			if (sm_is_err(ret)) {
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
					"sev=ERROR, func=%s, aq_rdq_rm=%m, flags=%#x", SMFCT,
					ret, aq_rcpt->aqr_flags);
				stopit = true;
				break;
			}

			sm_log_entry(qmgr_ctx->qmgr_lctx,
				QM_LCAT_SCHED, QM_LMOD_SCHED,
				SM_LOG_INFO, 9,
				SLT_SEV, "INFO",
				SLT_FUNC, "qmgr_sched_dlvry",
				SLT_SS_TA, aq_rcpt->aqr_ss_ta_id,
				SLT_DA_SE, dadb_entry->dadbe_da_se_id,
				SLT_DA_TA, dadb_entry->dadbe_da_ta_id,
				SLT_RCPT, aq_rcpt->aqr_pa,
				SLT_RCPT_IDX, aq_rcpt->aqr_idx,
				SLT_STATE, aq_rcpt->aqr_status,
				SLT(SLT_IP, 'I'), &AQR_CUR_DEST(aq_rcpt),
				SLT_I, i,
				SLT_REUSE, se_reuse,
				SLT_KEEP_OPEN, se_keep_open,
				SLT_E_CUR, qsc_ctx->qsc_dadb_ctx->dadb_entries_cur,
				SLT_E_MAX, qsc_ctx->qsc_dadb_ctx->dadb_entries_max,
				SLT_END);

			/* Add more recipients to this DA transaction? */
			for (aq_rcpt_nxt = AQR_SS_SUCC(aq_rcpt), nrcpts = 1;
			     aq_rcpt_nxt != aq_rcpt && nrcpts < max_rcpts_ta
			     && !AQ_TA_IS_FLAG(aq_ta, AQ_TA_FL_VERP)
			     && !AQR_IS_FLAG(aq_rcpt, AQR_FL_HAS_VERP);
			     aq_rcpt_nxt = AQR_SS_SUCC(aq_rcpt_nxt))
			{
/*
XXX "append" also undeliverable recipients
so they are counted in qda...
*/
				/* check whether DA and host are the same */
				if (AQR_SCHEDULE(aq_rcpt_nxt)
				    && SAME_TRANSACTION(aq_rcpt, aq_rcpt_nxt))
				{
					QM_LEV_DPRINTFC(QDC_SCHED, 1, (QM_DEBFP, "sev=DBG, func=%s, status=more, i=%u, ta=%s, rcpt=%@S, idx=%d, ip=%I, rcpt=%p, rcpt_nxt=%p, nrcpts=%u\n", SMFCT, i, aq_rcpt->aqr_ss_ta_id, aq_rcpt_nxt->aqr_pa, aq_rcpt_nxt->aqr_idx, &AQR_CUR_DEST(aq_rcpt), aq_rcpt, aq_rcpt_nxt, nrcpts));

					ret = qm_to_sc_add_rcpt(qsc_ctx, aq_rcpt_nxt, rcbe);
					if (sm_is_err(ret)) {
						sm_log_write(qmgr_ctx->qmgr_lctx,
							QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
							"sev=ERROR, func=%s, qm_to_sc_add_rcpt=%m", SMFCT,
							ret);
						stopit = true;
						break;
					}
					sm_log_entry(qmgr_ctx->qmgr_lctx,
						QM_LCAT_SCHED, QM_LMOD_SCHED,
						SM_LOG_INFO, 9,
						SLT_SEV, "INFO",
						SLT_FUNC, SMFCT,
						SLT_FREE, "sched", "more",
						SLT_SS_TA, aq_rcpt->aqr_ss_ta_id,
						SLT_DA_TA, dadb_entry->dadbe_da_ta_id,
						SLT_RCPT, aq_rcpt_nxt->aqr_pa,
						SLT_RCPT_IDX, aq_rcpt_nxt->aqr_idx,
						SLT(SLT_IP, 'I'), &AQR_CUR_DEST(aq_rcpt),
						SLT_END);

					AQR_SET_FLAG(aq_rcpt_nxt, AQR_FL_SCHED);
					SESSTA_COPY(aq_rcpt_nxt->aqr_da_ta_id, dadb_entry->dadbe_da_ta_id);

					/*
					**  Prepend new recipient to original one
					**  to keep the correct order.
					*/

					AQR_DA_PRE(aq_rcpt, aq_rcpt_nxt);
					++aq_ctx->aq_t_da;
					++nrcpts;

					ret = aq_rdq_rm(aq_ctx, aq_rcpt_nxt, THR_NO_LOCK, qmgr_ctx->qmgr_lctx);
					if (sm_is_err(ret)) {
						sm_log_write(qmgr_ctx->qmgr_lctx,
							QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
							"sev=ERROR, func=%s, aq_rdq_rm=%m, flags=%#x, where=more_rcpts2busy", SMFCT,
							ret, aq_rcpt_nxt->aqr_flags);
						stopit = true;
						break;
					}
				}

				/*
				**  Set this here??  That means it is only set if sent to a DA,
				**  not when it is "looked at" by the scheduler.
				**  Hence if the address doesn't resolve, last_try is never set.
				*/

				aq_rcpt_nxt->aqr_last_try = time_now;
			}

			/* Send task to DA */
			ret = sm_rcbcom_endrep(&qsc_ctx->qsc_com, qsc_ctx->qsc_com.rcbcom_tsk, true /* XXX HACK */, &rcbe);
			if (sm_is_err(ret)) {
				aq_rcpt_P aq_rcpt_r;

				if (rcbe != NULL) {
					/*
					**  rcbe has not been added to write list
					**  hence it needs to be cleaned up
					*/

					aq_rsnd_ctx_free(aq_rsnd_ctx);
					aq_rsnd_ctx = NULL;
					sm_rcbe_free(rcbe);
					rcbe = NULL;
				}

				/* undo status change for added recipients */
				for (aq_rcpt_nxt = AQR_DA_PRED(aq_rcpt);
				     aq_rcpt_nxt != aq_rcpt;
				     aq_rcpt_nxt = aq_rcpt_r)
				{
					aq_rcpt_r = AQR_DA_PRED(aq_rcpt);
					AQR_CLR_FLAG(aq_rcpt_nxt, AQR_FL_SCHED);

					res = aq_rdq_add(aq_ctx, aq_rcpt_nxt, NULL /* &aqrdq_flags */, THR_NO_LOCK);
					if (sm_is_err(res)) {
						sm_log_write(qmgr_ctx->qmgr_lctx,
							QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
							"sev=ERROR, func=%s, aq_rdq_add=%m, flags=%#x, where=more_rcpts2todo", SMFCT,
							ret, aq_rcpt_nxt->aqr_flags);
					}

					/* XXX also remove entry from list? */
					AQR_DA_DELENTRY(aq_rcpt_nxt);

					/*
					**  XXX Other things to undo?
					**  Make a list of things "done"
					**  so they can be "undone" here.
					*/

				}
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
					"sev=ERROR, func=%s, sm_qsc_endrep=%m", SMFCT, ret);
				stopit = true;
				break;
			}
			else
				aq_rsnd_ctx = NULL;

			/*
			**  At this point the information about the session/
			**  transaction is in the RCB list for the DA, but
			**  it isn't sent yet. Hence AQR_FL_WAIT4UPD can't
			**  be set here, it will be set when the RCB
			**  is sent to the DA.
			*/

			AQR_SET_FLAG(aq_rcpt, AQR_FL_SCHED);
			aq_rcpt->aqr_last_try = time_now;
			*pqsc_bits |= qsc_bit;
		}

		/*
		**  If stopit is set then the rdq should be rotated so
		**  the next time the scheduler starts at the one that
		**  couldn't be finished this time. XXX
		**  See begin of outer loop, it's currently rotated for
		**  every entry.
		*/

		if (rcbe != NULL) {
			/*
			**  Possible leftover from previous iteration.
			**  Maybe it should be reused?
			*/

			(void) aq_rsnd_ctx_free(aq_rsnd_ctx);
			aq_rsnd_ctx = NULL;
			sm_rcbe_free(rcbe);
			rcbe = NULL;
		}
		SM_ASSERT(NULL == aq_rsnd_ctx);
	}
	res = ret;

	/*
	**  XXX Crude resource control...
	**  This needs to be more sophisticated and take more resources
	**  into account. Moreover, it needs to free memory if required,
	**  and it needs to change the limits on other data structures
	**  (esp. IQDB) too.
	*/

	if (sm_is_err(res) && sm_error_value(res) == ENOMEM) {
#if 0
		/* throttle servers; do this in caller?? */
		(void) qm_control(qmgr_ctx, 1, 100, QMGR_RFL_MEM_I, THR_LOCK_UNLOCK);
#endif
		ret = aq_lower_limit(aq_ctx, 0, THR_NO_LOCK);
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_WARN, 8,
			"sev=WARN, func=%s, aq_lower_limit=%m", SMFCT, ret);
	}
	else
	{
		/*
		**  XXX This should NOT be done always... but as long as we don't
		**  have a way to measure memory usage...
		**  How about a way to gradually increase it, similar to qss_control()?
		*/

		ret = aq_raise_limit(aq_ctx, UINT_MAX, THR_NO_LOCK);
	}

	if (rcbe != NULL) {
		aq_rsnd_ctx_free(aq_rsnd_ctx);
		aq_rsnd_ctx = NULL;
		sm_rcbe_free(rcbe);
		rcbe = NULL;
	}
	SM_ASSERT(NULL == aq_rsnd_ctx);
	return res;
}

/*
**  QMGR_SCHED -- scheduler
**	this is running as a task (timeout activated)
**
**	Parameters:
**		tsk -- evthr task
**
**	Returns:
**		usual sm_error code
**
**	Locking: locks entire aq_ctx during operation, returns unlocked
**
**	Last code review:
**	Last code change:
*/

sm_ret_T
qmgr_sched(sm_evthr_task_P tsk)
{
	int r;
	sm_ret_T ret;
	qmgr_ctx_P qmgr_ctx;
	aq_ctx_P aq_ctx;
	bool err;
	uint32_t qsc_bits;
	timeval_T timeval_now, delay;
	int delay_next_try;
#if QMGR_TEST
	int delay_next_try_tst;
#endif

	SM_IS_EVTHR_TSK(tsk);
	qmgr_ctx = (qmgr_ctx_P) tsk->evthr_t_actx;
	SM_IS_QMGR_CTX(qmgr_ctx);
	aq_ctx = qmgr_ctx->qmgr_aq;
	SM_IS_AQ(aq_ctx);
	err = false;
	qsc_bits = 0;

	/* Check ret?? */
	ret = evthr_timeval(tsk->evthr_t_ctx, &timeval_now);
	delay_next_try = 0;

#if QMGR_TEST
	delay_next_try_tst = 0;
	if (QCNF_IS_FLAG(qmgr_ctx, QCNF_FL_NO_SCHED)) {
		delay.tv_usec = 0;
		delay.tv_sec = 100;
		timeradd(&timeval_now, &delay, &tsk->evthr_t_sleep);
		return EVTHR_SLPQ;
	}

	if (SM_IS_FLAG(qmgr_ctx->qmgr_cnf.q_cnf_tests, QMGR_TEST_INT_SRC)) {
		ret = qm_tst_fill_aq(qmgr_ctx);
		if (sm_is_err(ret))
			delay_next_try_tst = sm_is_warn_err(ret) ?
			                     (0 - sm_error_value(ret)) : 300;
		else if (SM_NOTDONE == ret)
			delay_next_try_tst = 1;
# if QMGR_STATS
		if (qmgr_ctx->qmgr_cnf.q_cnf_tot_tas <= qmgr_ctx->qmgr_tas_sent)
sm_io_fprintf(smioerr, "q_cnf_tot_tas=%u, tas_sent=%lu, aq_entries=%u\n", qmgr_ctx->qmgr_cnf.q_cnf_tot_tas, qmgr_ctx->qmgr_tas_sent, aq_ctx->aq_entries);
		if (qmgr_ctx->qmgr_cnf.q_cnf_tot_tas <= qmgr_ctx->qmgr_tas_sent
		    && aq_ctx->aq_entries <= 2)
		{
			qm_info(qmgr_ctx, smioout);
			return EVTHR_TERM;
		}
# endif
	}

	if (SM_IS_FLAG(qmgr_ctx->qmgr_cnf.q_cnf_tests, QMGR_TEST_SCHED_DLY)) {
		static time_T sched_start_delay = 0;

		if (0 == sched_start_delay)
			sched_start_delay = timeval_now.tv_sec + 10;
QM_LEV_DPRINTFC(QDC_SCHED, 6, (QM_DEBFP, "sev=DBG, func=qmgr_sched, sched_start_delay=%ld, timeval_now.tv_sec=%ld\n", sched_start_delay, timeval_now.tv_sec));
		if (sched_start_delay > timeval_now.tv_sec) {
			tsk->evthr_t_sleep.tv_sec = sched_start_delay;
			return EVTHR_SLPQ;
		}
		SM_CLR_FLAG(qmgr_ctx->qmgr_cnf.q_cnf_tests, QMGR_TEST_SCHED_DLY);
	}
#endif /* QMGR_TEST */

	ret = qm_get_edb_entries(qmgr_ctx, &delay_next_try);
	if (delay_next_try < 0)
		delay_next_try = 1;	/* XXX HACK */

	/* HACK notify qar task */
	if (QDA_ACT_SMAR(ret)) {
		sm_evthr_task_P qar_tsk;
		qar_ctx_P qar_ctx;

		qar_ctx = qmgr_ctx->qmgr_ar_ctx;
		SM_IS_QAR_CTX(qar_ctx);
		qar_tsk = qmgr_ctx->qmgr_ar_tsk;
		if (qar_tsk != NULL)
			ret = evthr_en_wr(qar_tsk);
		else
			ret = sm_error_temp(SM_EM_Q_SCHED, SM_E_NO_AR);
	}

	QM_LEV_DPRINTFC(QDC_SCHED, 6, (QM_DEBFP, "sev=DBG, func=qmgr_sched, tsk=%p\n", tsk));
	r = pthread_mutex_lock(&aq_ctx->aq_mutex);
	SM_LOCK_OK(r);
	if (r != 0) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_CRIT, 4,
			"sev=CRIT, func=qmgr_sched, lock_aq=%d", r);
		goto error;
	}

	/* XXX just wait till someone wakes this up? */

	if (aq_ctx->aq_nextrun.tv_sec != 0) {
		if (timercmp(&timeval_now, &aq_ctx->aq_nextrun, <)) {
			tsk->evthr_t_sleep = aq_ctx->aq_nextrun;
			QM_LEV_DPRINTFC(QDC_SCHED, 2, (QM_DEBFP, "sev=DBG, func=qmgr_sched, delay=forced, next_try=%ld\n", aq_ctx->aq_nextrun.tv_sec));
			goto unlock;
		}

		aq_ctx->aq_nextrun.tv_usec = 0;
		aq_ctx->aq_nextrun.tv_sec = 0;
	}

	if (aq_is_empty(aq_ctx)) {
		delay.tv_usec = 0;
		delay.tv_sec = delay_next_try == 0 ? 3000 : delay_next_try;
		QM_LEV_DPRINTFC(QDC_SCHED, 2, (QM_DEBFP, "sev=DBG, func=qmgr_sched, aq=empty, next_try=%d, delay=%ld\n", delay_next_try, delay.tv_sec));
	}
	else {
		/* do something.... */
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_DEBUG, 14,
			"sev=DBG, func=qmgr_sched, aq_entries=%u, aq_t_da=%u",
			aq_ctx->aq_entries, aq_ctx->aq_t_da);
		ret = qmgr_sched_dlvry(qmgr_ctx, aq_ctx, &qsc_bits, &delay_next_try);
		if (sm_is_err(ret)) {
			/* XXX Map error type to logging category and sev? */
			err = sm_is_error(ret);
			sm_log_write(qmgr_ctx->qmgr_lctx,
				QM_LCAT_SCHED, QM_LMOD_SCHED,
				err ? SM_LOG_ERR : SM_LOG_WARN,
				err ? 2 : QMGR_IS_SFLAG(qmgr_ctx, QMGR_SFL_DA) ? 14 : 12,
				"sev=%s, func=qmgr_sched, qmgr_sched_dlvry=%m",
				err ? "ERROR" : "WARN",
				ret);
			if (sm_error_warn(SM_EM_Q_SCHED, SM_E_NO_DA) == ret &&
				(r = aq_usage(aq_ctx, AQ_USAGE_ALL))
				 >= qmgr_ctx->qmgr_upper[QMGR_RFL_AQ])
			{
				QMGR_SET_SFLAG(qmgr_ctx, QMGR_SFL_RSTD);
				return EVTHR_TERM;
			}
		}

		/* XXX interval... */
		delay.tv_usec = 0;
#if QMGR_TEST
		if (SM_IS_FLAG(qmgr_ctx->qmgr_cnf.q_cnf_tests, QMGR_TEST_INT_SRC)) {
			r = aq_usage(aq_ctx, AQ_USAGE_ALL);
			if (r < qmgr_ctx->qmgr_cnf.q_cnf_max_fill_aq &&
			    0 == delay_next_try_tst)
			{
				delay.tv_sec = 0;
				delay.tv_usec = r;
			}
			else if (delay_next_try_tst < 0) {
				delay.tv_sec = 0;
				delay.tv_usec = (-delay_next_try_tst) * 500;
			}
			else
				delay.tv_sec = delay_next_try_tst == 0 ? 1 : delay_next_try_tst;
/* sm_io_fprintf(smioerr, "usage=%d, delay_next_try_tst=%d, delay_next_try=%d, delay=%d.%06ld\n", r, delay_next_try_tst, delay_next_try, delay.tv_sec, delay.tv_usec); */
		}
		else
#endif
		delay.tv_sec = delay_next_try == 0 ? 600 : delay_next_try;
		QM_LEV_DPRINTFC(QDC_SCHED, 3, (QM_DEBFP, "sev=DBG, func=qmgr_sched, next_try=%d, delay=%ld\n", delay_next_try, delay.tv_sec));
	}

	QM_LEV_DPRINTFC(QDC_SCHED, 4, (QM_DEBFP, "sev=DBG, func=qmgr_sched, old t_sleep=%ld\n", tsk->evthr_t_sleep.tv_sec));
	timeradd(&timeval_now, &delay, &tsk->evthr_t_sleep);
	QM_LEV_DPRINTFC(QDC_SCHED, 4, (QM_DEBFP, "sev=DBG, func=qmgr_sched, timeval_now=%ld, delay=%ld, t_sleep=%ld\n", timeval_now.tv_sec, delay.tv_sec, tsk->evthr_t_sleep.tv_sec));
	if (err)
		aq_ctx->aq_nextrun = tsk->evthr_t_sleep;

  unlock:
	r = pthread_mutex_unlock(&aq_ctx->aq_mutex);
	if (r != 0) {
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
			"sev=ERROR, func=qmgr_sched, unlock_aq=%d", r);
		goto error;
	}

	/*
	**  notify qsc task after aq has been unlocked, otherwise
	**  there might be a deadlock (qsc task locked, waiting for aq,
	**  while this task has the lock for aq and tries to notify
	**  (which requires locking) the qsc task).
	*/

	if (qsc_bits != 0) {
		uint i;
		uint32_t j;
		qsc_ctx_P qsc_ctx;

		r = pthread_mutex_lock(&qmgr_ctx->qmgr_mutex);
		SM_LOCK_OK(r);
		if (r != 0) {
			sm_log_write(qmgr_ctx->qmgr_lctx,
				QM_LCAT_SMTPC, QM_LMOD_FROM_SMTPC,
				SM_LOG_CRIT, 4,
				"sev=CRIT, func=qmgr_sched, lock_qmgr=%d",
				r);
			goto error;
		}

		/* XXX HACK! Must match what qmgr_sched_dlvry does! */
		for (j = 1, i = 0; i < QM_N_SC_GLI(qmgr_ctx); i++, j *= 2) {
			if ((qsc_bits & j) == 0)
				continue;

			qsc_ctx = qmgr_li_sc(qmgr_ctx, i);
			if (qsc_ctx->qsc_com.rcbcom_tsk == NULL)
				continue;
			SM_IS_EVTHR_TSK(qsc_ctx->qsc_com.rcbcom_tsk);

			/* tell someone to send the tasks to the DAs */
			ret = evthr_en_wr(qsc_ctx->qsc_com.rcbcom_tsk);
			if (sm_is_err(ret)) {
				sm_log_write(qmgr_ctx->qmgr_lctx,
					QM_LCAT_SCHED, QM_LMOD_SCHED, SM_LOG_ERR, 1,
					"sev=ERROR, func=qmgr_sched, evthr_en_wr=%m", ret);
				/* XXX what now my friend? */
			}
		}
		r = pthread_mutex_unlock(&qmgr_ctx->qmgr_mutex);
		SM_ASSERT(0 == r);
	}

	return EVTHR_SLPQ;

  error:
	delay.tv_usec = 0;
	delay.tv_sec = 1;
	timeradd(&timeval_now, &delay, &tsk->evthr_t_sleep);
	return EVTHR_SLPQ;
}
